# UTF-8
# Claudio Perez
# ema


import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import copy
from emma.elements import *
# import tensorflow as tf


class Model:
    def __init__(self, ndm, ndf):
        """Basic model object
        
        Parameters
        -----------
        ndm:
            number of model dimensions
        ndf:
            number of degrees of freedom (dofs) at each node

        """
        self.ndf: int = ndf
        self.ndm: int = ndm
        self.DOF: list = None
        self.dtype='float32'

        # Define DOF list indexing 
        if ndm == 1:
            self.prob_type = '1d'
            self.ddof: dict = {'x': 0}  # Degrees of freedom at each node
        
        if ndf == 2:
            self.prob_type = '2d-truss'
            self.ddof: dict = { 'x': 0, 'y': 1} # Degrees of freedom
        elif ndm == 2 and ndf ==3:
            self.prob_type = '2d-frame'
            self.ddof: dict = { 'x': 0, 'y': 1, 'rz':2}
        elif ndm == 3 and ndf ==3:
            self.prob_type = '3d-truss'
            self.ddof: dict = { 'x': 0, 'y': 1, 'z':2}
        elif ndm == 3 and ndf ==6:
            self.prob_type = '3d-frame'
            self.ddof: dict = { 'x': 0, 'y': 1, 'z':2, 'rx':3, 'ry':4, 'rz':5}
            
        # model inventory lists
        self.elems: list = []
        self.nodes: list = []
        self.rxns:  list = []
        self.hinges: list = []
        self.iforces: list = []
        self.loads: list = []
        self.states: list = []
        self.redundants: list = []

        # model inventory dictionaries
        self.delems: dict = {}
        self.dnodes: dict = {}
        self.dxsecs: dict = {}
        self.dhinges: dict = {}
        self.dstates: dict = {}
        self.dxsecs: dict = {}
        self.materials: dict = {}
        self.xsecs: dict = {}
        self.dredundants: dict = {}
        
        # Initialize default material/section properties
        self.material('default', 1.0)
        self.xsection('default', 1.0, 1.0)

    @property
    def rel(self):
        return [rel for elem in self.elems for rel in elem.rel.values()]

    @property
    def nn(self) -> int:
        """return number of nodes in model"""
        return len(self.nodes)
    
    @property
    def nr(self) -> int:
        """return number of constrained dofs in model"""
        return len(self.rxns)
    
    @property
    def ne(self) -> int:
        """return number of elements in model"""
        return len(self.elems)

    @property
    def nQ(self):
        f = 0
        for elem in self.elems:
            f += len(elem.basic_forces)
        return f

    @property
    def nq(self):
        f = []
        for elem in self.elems:
            f.append(sum([1 for x in elem.q]))
        return f
    
    @property
    def nv(self):
        lst = []
        for elem in self.elems:
            lst.append(sum([1 for x in elem.v]))
        return lst

    @property
    def nf(self) -> int:
        x = self.nt - self.nr
        return x

    @property
    def nt(self) -> int:
        return self.ndf*self.nn

    @property
    def rdofs(self): 
        """Return list of restrained dofs"""
        DOF = self.DOF

        return []

    @property
    def NOS(self) -> int:
        nf = self.nf
        nq = sum(self.nq)
        return nq - nf

    @property
    def basic_forces(self):
        return np.array([q for elem in self.elems for q in elem.basic_forces ])

    @property
    def rdnt_forces(self):
        cforces = self.cforces
        return np.array([q  for q in cforces if q.redundant ])
    
    @property
    def cforces(self):
        return np.array([q for elem in self.elems for q in elem.basic_forces if not q.rel])
    
    @property 
    def eforces(self):
        """Return array of elastic element forces"""
        return np.array([q for elem in self.elems for q in elem.basic_forces if (q.plastic_event is None)])

    @property
    def idx_c(self):
        cforces = self.cforces
        forces = self.basic_forces
        idx_c = np.where(np.isin(forces,cforces))[0]
        return idx_c

    @property
    def idx_e(self):
        """return indices of elastic basic (not plastic) forces"""
        cforces = self.cforces
        eforces = self.eforces
        idx_e = np.where(np.isin(cforces,eforces))[0]
        return idx_e
    
    @property
    def idx_f(self):
        return np.arange(0,self.nf)
    @property
    def idxx_f(self):
        return None
    
    @property
    def idx_i(self):
        rdts = self.rdnt_forces
        forces = self.basic_forces
        idx_i = np.where(np.logical_not(np.isin(forces, rdts)))[0]
        return idx_i

    @property
    def idx_x(self):
        rdts = self.rdnt_forces
        forces = self.basic_forces
        idx_x = np.where(np.isin(forces, rdts))[0]
        return idx_x

    def node(self, tag: str, x: float, y=None, z=None, mass: float=None):
        newNode = Node(self, tag, self.ndf, [x, y, z], mass)
        self.nodes.append(newNode)
        self.dnodes.update({newNode.tag : newNode})
        return newNode
    
    def state(self, method="Linear"):
        if self.DOF==None: self.numDOF()
        newState = State(self, method)

        ElemTypes = {type(elem) for elem in self.elems}
        StateVars = {key for elem in ElemTypes for key in elem.stateVars.keys() }

        stateDict = {var : {elem.tag : copy.deepcopy(elem.stateVars[var]) for elem in self.elems if var in elem.stateVars.keys()} for var in StateVars}
        self.states.append(stateDict)
        return stateDict
    
    def numDOF(self):
        crxns = self.ndf*len(self.nodes) - len(self.rxns)+1
        df = 1
        temp = []
        for node in self.nodes:
            DOFs = []
            for rxn in node.rxns:
                if not(rxn):
                    DOFs.append(df)
                    df += 1
                else:
                    DOFs.append(crxns)
                    crxns += 1
            temp.append(DOFs)
        self.DOF = temp
        return self.DOF

    def update(self,U_vector):
        for node in self.nodes:
            delta = [0.,0.]
            for i,dof in enumerate(node.dofs[0:2]):
                if not node.rxns[i]:
                    try: delta[i] = U_vector[U_vector.row_data.index(str(dof))]
                    except: pass
            
            node.xi= node.x
            node.x = delta[0] + node.xi
            node.yi= node.y
            node.y = delta[1] + node.yi
        pass

    def fix(self, node, dirn=["x","y","rz"]): # for dirn enter string (e.g. "x", 'y', 'rz')
        if isinstance(dirn,list):
            rxns = []
            for df in dirn:
                newRxn = Rxn(node, df)
                self.rxns.append(newRxn)
                rxns.append(newRxn)
                node.rxns[self.ddof[df]] = 1
            return rxns
        else:
            newRxn = Rxn(node, dirn)
            self.rxns.append(newRxn)
            node.rxns[self.ddof[dirn]] = 1
            return newRxn

    def boun(self, node, ones):
        for i, dof in enumerate(self.ddof):
            if ones[i]:
                self.fix(node, dof)

    def pin(self, node):
        """creates pinned reaction at specified node

        """
        self.fix(node, ['x', 'y'])
        if self.ndm == 3:
            self.fix(node, 'z')
        return

    def roller(self, node):
        """creates roller reaction at specified node

        """
        self.fix(node, 'y')
        return

 # Other
    def material(self, tag: str, E: float):
        newMat = Material(tag, E)
        self.materials[tag]=newMat
        return newMat

    def xsection(self, tag: str, A: float, I: float):
        newXSect = XSect(tag, A, I)
        self.xsecs[tag] = newXSect
        return newXSect

 # Elements
    def add_element(self, element):
        """Add a general element to model
        
        Parameters
        ---------
        element : obj

        """

        self.elems.append(element)
        self.delems.update({element.tag:element})

        for node in element.nodes:
            node.elems.append(element)
        
        return element

    def add_elements(self, elements):
        """Add a general element to model
        
        Parameters
        ---------
        element : obj

        """
        for element in elements:
            self.elems.append(element)
            self.delems.update({element.tag:element})
            for node in element.nodes:
                node.elems.append(element)
        
        return elements


    def beam(self, tag: str, iNode, jNode, mat=None, sec=None, Qpl=None,**kwds):
        """Add a 2D linear Euler-Bernouli beam object to model
        
        Parameters
        ---------
        tag : str
            string used for identifying object
        iNode : emma.Node
            node object at element i-end
        jNode : emma.Node
            node object at element j-end
        mat : emma.Material 

        sec : emma.Section
        

        """

        if mat is None:
            E = kwds["E"] if "E" in kwds else self.materials["default"].E
        else:
            E = mat.E

        if sec is None:
            A = kwds["A"] if "A" in kwds else self.xsecs["default"].A
            I = kwds["I"] if "I" in kwds else self.xsecs["default"].I
        else:
            A = sec.A
            I = sec.I


        newElem = Beam(tag, iNode, jNode, E, A, I)
        self.elems.append(newElem)
        self.delems.update({newElem.tag:newElem})
        # self.connect([iNode, jNode], "Beam") # considering deprecation
        iNode.elems.append(newElem)
        jNode.elems.append(newElem)

        if Qpl is not None:
            newElem.Qpl = np.zeros((3,2))
            newElem.Np = [Qpl[0], Qpl[0]]
            newElem.Mp = [[Qpl[1], Qpl[1]],[Qpl[2], Qpl[2]]]
            for i, key in enumerate(newElem.Qp['+']):
                newElem.Qp['+'][key] = newElem.Qp['-'][key] = Qpl[i] # consider depraction of elem.Qp in favor of elem.Qpl
                newElem.Qp['+'][key] = newElem.Qp['-'][key] = Qpl[i]
                newElem.Qpl[i,:] = Qpl[i] # <- consider shifting to this format for storing plastic capacities
        return newElem
    
    def tensor_beam(self, tag: str, iNode, jNode, mat=None, sec=None, Qpl=None):
        """Add a 2D linear Euler-Bernouli beam object to model
        
        Parameters
        ---------
        tag : str
            string used for identifying object
        iNode : emma.Node
            node object at element i-end
        jNode : emma.Node
            node object at element j-end
        mat : emma.Material 

        sec : emma.Section
        

        """

        if mat is None:
            mat = self.materials['default']
        if sec is None:
            sec = self.xsecs['default']
        newElem = TensorBeam(tag, iNode, jNode, mat.E, sec.A, sec.I)
        self.elems.append(newElem)
        self.delems.update({newElem.tag:newElem})
        # self.connect([iNode, jNode], "Beam") # considering deprecation
        iNode.elems.append(newElem)
        jNode.elems.append(newElem)
        
        if Qpl is not None:
            newElem.Qpl = np.zeros((3,2))
            newElem.Np = [Qpl[0], Qpl[0]]
            newElem.Mp = [[Qpl[1], Qpl[1]],[Qpl[2], Qpl[2]]]
            for i, key in enumerate(newElem.Qp['+']):
                newElem.Qp['+'][key] = newElem.Qp['-'][key] = Qpl[i] # consider depraction of elem.Qp in favor of elem.Qpl
                newElem.Qp['+'][key] = newElem.Qp['-'][key] = Qpl[i]
                newElem.Qpl[i,:] = Qpl[i] # <- consider shifting to this format for storing plastic capacities
        return newElem

    def girder(self, nodes, mats=None, xsecs=None, story=None):
        tags=[chr(i) for i in range(ord("a"), ord("a") + 26)]

        if mats is None: mats = [self.materials['default']]*(len(nodes)-1)
        if xsecs is None: xsecs = [self.xsecs['default']]*(len(nodes)-1)
        newElems = []

        for i, nd in enumerate(nodes[0:len(nodes)-1]):
            iNode = nd
            jNode = nodes[i+1]
            if story is None: tag = tags[i]
            else: tag = tags[story]+str(len(newElems)+1)
            newElem = Beam(tag, iNode, jNode, mats[i].E, xsecs[i].A, xsecs[i].I)
            self.elems.append(newElem)
            self.delems.update({newElem.tag:newElem})
            iNode.elems.append(newElem)
            jNode.elems.append(newElem)
            newElems.append(newElem)
        return newElems

    def frame(self, bays, stories, column_mat=None, column_sec=None, 
                                   girder_mat=None, girder_sec=None):
        """Macro for generating rectangular building frames
        
        Parameters
        ---------------------
        bays: tuple
            tuple containing bay width, and number of bays
        stories: tuple
            tuple
        column_mat: 
        
        """
        o = {'x':0.0, 'y': 0.0}
        w = bays[1]     # Bay width
        nb = bays[0]    # Number of bays
        h = stories[1]  # Story height
        ns = stories[0] # Number of stories

        if girder_mat==None: girder_mat = self.materials['default']
        if girder_sec==None: girder_sec = self.xsecs['default']
        if column_mat==None: column_mat = self.materials['default']
        if column_sec==None: column_sec = self.xsecs['default']

        self.snodes = []
        ntag = 1 # Counter for node tags
        ctag = 1 # Counter for column tags
        for s in range(ns+1):
            snodes = []
            for b in range(nb+1):
                snodes.append(self.node(str(ntag), o['x']+b*w, o['y']+s*h))
                ntag += 1
                if not s == 0:
                    self.beam('cl' + str(ctag), self.snodes[s-1][b], snodes[b], column_mat, column_sec)
                    ctag += 1
            if not s == 0: self.girder(snodes, [girder_mat]*nb, [girder_sec]*nb, s-1)
            self.snodes.append(snodes)

    def truss(self, tag: str, iNode, jNode, mat=None, xsec=None, Qpl=None,A=None,E=None):
        if mat is None: mat = self.materials['default']
        if E is None: E = mat.E
        # cross section
        if xsec is None: xsec = self.xsecs['default']
        if A is None: A = xsec.A

        newElem = Truss(tag, iNode, jNode, E, A)
        self.delems.update({newElem.tag:newElem})
        self.elems.append(newElem)
        iNode.elems.append(newElem)
        jNode.elems.append(newElem)

        if Qpl is not None:
            newElem.Np = [Qpl[0], Qpl[0]]
            newElem.Qp['+']['1'] = newElem.Qp['-']['1'] = Qpl[0]
        return newElem
    
    def tensor_truss(self, tag: str, iNode, jNode, mat=None, xsec=None, Qpl=None,A=None,E=None):
        if mat is None: mat = self.materials['default']
        if E is None: E = mat.E
        # cross section
        if xsec is None: xsec = self.xsecs['default']
        if A is None: A = xsec.A

        newElem = TensorTruss(tag, iNode, jNode, E, A)
        self.delems.update({newElem.tag:newElem})
        self.elems.append(newElem)
        iNode.elems.append(newElem)
        jNode.elems.append(newElem)

        if Qpl is not None:
            newElem.Np = [Qpl[0], Qpl[0]]
            newElem.Qp['+']['1'] = newElem.Qp['-']['1'] = Qpl[0]
        return newElem

    def taprod(self, tag: str, iNode, jNode, mat=None, xsec=None, Qpl=None,A=None,E=None):
        """Construct a tapered rod element with variable E and A values."""
        if mat is None: mat = self.materials['default']
        if E is None: E = mat.E
        # cross section
        if xsec is None: xsec = self.xsecs['default']
        if A is None: A = xsec.A
        
        newElem = TaperedTruss(tag, iNode, jNode, E, A)
        self.delems.update({newElem.tag:newElem})
        self.elems.append(newElem)
        iNode.elems.append(newElem)
        jNode.elems.append(newElem)

        if Qpl is not None:
            newElem.Np = [Qpl[0], Qpl[0]]
            newElem.Qp['+']['1'] = newElem.Qp['-']['1'] = Qpl[0]
        return newElem
    
    def truss3d(self, tag: str, iNode, jNode, mat=None, xsec=None):
        """Add an emma.Truss3d object to model
        
        Parameters
        ---------

        """
        if mat is None: mat = self.materials['default']
        if xsec is None: xsec = self.xsecs['default']
        newElem = Truss3D(tag, iNode, jNode, mat, xsec)
        self.elems.append(newElem)
        self.delems.update({newElem.tag:newElem})
        iNode.elems.append(newElem)
        jNode.elems.append(newElem)
        return newElem

    def hinge(self, elem, node): # pin a beam end.
        newHinge = Hinge(elem, node)
        self.hinges.append(newHinge)
        if node == elem.nodes[0]:
            elem.rel['2'] = True
            elem.q.pop('2')
            elem.basic_forces[1].rel = True
        elif node == elem.nodes[1]:
            elem.rel['3'] = True
            elem.q.pop('3')
            elem.basic_forces[2].rel = True
        else: print("Error: element {} is not bound to node {}.".format(elem, node))

        return newHinge

    def redundant(self, elem: object, nature):
        """
        nature:
        """

        newq = IntForce(elem, nature, nature)
        elem.red[nature] = True
        self.redundants.append(newq)

    def update(self, U):
        for node in self.nodes:
            delta = [0.,0.]
            for i, dof in enumerate(node.dofs[0:2]):
                if not node.rxns[i]:
                    try:
                        delta[i] = U[U.row_data.index(str(dof))]
                    except: pass
                node.x += delta[0]
                node.y += delta[1]

    def load_state(self, state):
        for elem in state["elems"]:
            if "q" in state["elems"][elem]:
                for i, q in enumerate(self.delems[elem].q):
                    self.delems[elem].q[q] = state["elems"][elem]["q"][i]


class rModel(Model):
    def __init__(self, ndm, ndf):
        super().__init__(ndm=2, ndf=3)
        self.material('default', 1.0)
        self.xsection('default', 1.0, 1.0)
    
    def isortho(self, elem):
        if (abs(elem.cs) == 1.0) or (abs(elem.sn) == 1.0):
            return True
        else:
            return False

    def numdofs(self):
        current_rxn = 1
        current_dof = 1
        rxn_ixs = []
        DOFs = [[0, 0, 0] for node in self.nodes]
        for i, node in enumerate(self.nodes):
            # x-dof
            dirn = 0
            if not(node.rxns[dirn]): # node is free
                if not(DOFs[i][dirn]): # node unassigned
                    for elem in node.elems:
                        if abs(elem.cs) == 1.0: # x-dof coupled to far end
                            if elem.nodes[0] == node:
                                far_node = self.nodes.index(elem.nodes[1])
                            if elem.nodes[1] == node:
                                far_node = self.nodes.index(elem.nodes[0])

                            if not(DOFs[far_node][dirn]): # Far node dof unassigned
                                if not(self.nodes[far_node].rxns[dirn]): # Far node is free
                                    DOFs[far_node][dirn] = current_dof
                                    DOFs[i][dirn] = current_dof
                                    current_dof += 1
                                else: # Far node is fixed
                                    DOFs[far_node][dirn] = current_rxn
                                    DOFs[i][dirn] = current_rxn
                                    current_rxn += 1
                                    rxn_ixs.append( (i,dirn) )
                                    rxn_ixs.append( (far_node,dirn) )
                            else: # Far node dof already assigned
                                DOFs[i][dirn] = DOFs[far_node][dirn]
                                if self.nodes[far_node].rxns[dirn]: # Far node is fixed
                                    rxn_ixs.append( (i,dirn) )
                        elif all([abs(elem.cs) != 1.0 for elem in node.elems]): # x-dof free/uncoupled
                            if not(DOFs[i][dirn]):
                                DOFs[i][dirn] = current_dof 
                                current_dof += 1
            else: # node is fixed
                if not(DOFs[i][dirn]): # node is unassigned
                    DOFs[i][dirn] = current_rxn 
                    current_rxn += 1
                    rxn_ixs.append( (i,dirn) )

            # y-dof
            dirn = 1
            if not(node.rxns[dirn]):
                if not(DOFs[i][dirn]):
                    for elem in node.elems:
                        if abs(elem.sn) == 1.0:
                            # get far node index
                            if elem.nodes[0] == node:
                                far_node = self.nodes.index(elem.nodes[1])
                            if elem.nodes[1] == node:
                                far_node = self.nodes.index(elem.nodes[0])
                            
                            if not(DOFs[far_node][dirn]):
                                if not(self.nodes[far_node].rxns[dirn]):
                                    DOFs[far_node][dirn] = current_dof
                                    DOFs[i][dirn] = current_dof
                                    current_dof += 1
                                else:
                                    DOFs[far_node][dirn] = current_rxn
                                    DOFs[i][dirn] = current_rxn
                                    current_rxn += 1
                                    rxn_ixs.append( (i,dirn) )
                                    rxn_ixs.append( (far_node,dirn) )
                            else: 
                                DOFs[i][dirn] = DOFs[far_node][dirn]
                                if self.nodes[far_node].rxns[dirn]:
                                    rxn_ixs.append( (i,dirn) )
                        elif all([abs(elem.sn) != 1.0 for elem in node.elems]):
                            if not(DOFs[i][dirn]):
                                DOFs[i][dirn] = current_dof
                                current_dof += 1
            else:
                if not(DOFs[i][dirn]):
                    DOFs[i][dirn] = current_rxn
                    current_rxn += 1
                    rxn_ixs.append( (i,dirn) )
            
          # rz-dof
            dirn = 2
            if not(node.rxns[2]):
                DOFs[i][dirn] = current_dof 
                current_dof += 1
            else:
                DOFs[i][dirn] = current_rxn 
                current_rxn += 1
                rxn_ixs.append( (i,dirn) )

        for ids in rxn_ixs:
            DOFs[ids[0]][ids[1]] += current_dof - 1

        return DOFs
         
    
    def numDOF(self):
        crxns = self.ndf*len(self.nodes) - len(self.rxns)+1
        df = 1
        temp = []
        for node in self.nodes:
            DOFs = []
            for rxn in node.rxns:
                if not(rxn):
                    DOFs.append(df)
                    df += 1
                else:
                    DOFs.append(crxns)
                    crxns += 1
            temp.append(DOFs)
        self.DOF = temp
        return self.DOF

    @property 
    def triv_forces(self):
        """list of trivial axial forces"""
        lst = []
        for elem in self.elems:
            if len(elem.basic_forces) > 1:
                if elem.dofs[0]==elem.dofs[3] or elem.dofs[1]==elem.dofs[4]:
                    lst.append(elem.basic_forces[0])
        return np.array(lst)

    # @property
    # def basic_forces(self):
    #     # bmax = self.TrAx_forces
    #     forces = np.array([q for elem in self.elems for q in elem.basic_forces if not q.rel])
    #     return forces
    
    @property
    def cforces(self):
        triv = self.triv_forces
        arry = np.array([q for elem in self.elems for q in elem.basic_forces if (
            (q.plastic_event is None) and (
            not q in triv) and (
            not q.rel))])
        return arry

    @property
    def nr(self):
        return len(self.rxns)

    # @property
    # def nq(self):
    #     f = []
    #     for elem in self.elems:
    #         f.append(sum([1 for q in elem.basic_forces if not q.rel and (not q in self.triv_forces)]))
    #     return f
    
    @property
    def nv(self):
        """Returns number of element deformations in model"""
        lst = []
        for elem in self.elems:
            lst.append(sum([1 for x in elem.v]))
        return lst


    
    @property
    def fdof(self): 
        """Return list of free dofs"""
        pass

    @property
    def nt(self):
        nt = max([max(dof) for dof in self.DOF])
        return nt
    
    @property
    def nm(self):
        """No. of kinematic mechanisms, or the no. of dimensions spanned by the 
        m independent inextensional mechanisms.
        
        """
        # A = A_matrix(mdl)
        pass

    @property
    def NOS(self):
        nf = self.nf
        nq = sum(self.nq)
        return nq - nf


class DofSpace:

    def __init__ ( self, Model ):
        self.ndf = Model.ndf
        self.nn = Model.nn
        # self.dofTypes = elements.getDofTypes()
        self.dofs     = np.array( range( self.nn * self.ndf ) ).reshape( ( self.nn, self.ndf ) )
        self.model = Model

        #Create the ID map
        # self.IDmap = itemList()
        self.IDmap = {}

        for ind,node in enumerate(Model.nodes):
            if node.tag in self.IDmap:
                raise RuntimeError( 'ID ' + str(node.tag) + ' already exists in ' + type(self.IDmap).__name__ )
            self.IDmap[node.tag] = ind

        self.constrainedDofs = []
        self.constrainedVals = []
        self.constrainedFac  = 1.0

    def __str__ ( self ):
        return str(self.model.dofs)

    def __len__ ( self ):
        return self.model.nt

    def setConstrainFactor( self , fac ):
        self.constrainedFac = fac
    
    def readFromFile( self, fname ):
        pass

    def constrain ( self, nodeID, dofTypes , val = 0. ):
        if not nodeID in self.model.nodes:
            raise RuntimeError('Node ID ' + str(nodeID) + ' does not exist')

        ind = self.IDmap.get( nodeID )

        if isinstance( dofTypes, str ):
            dofTypes = [dofTypes]

        #Check if the dofTypes exist
        for dofType in dofTypes:
            if dofType not in self.dofTypes:
                raise RuntimeError('DOF type "' + dofType + '" does not exist')
            
        for dofType in dofTypes:    
            self.constrainedDofs.append( self.dofs[ind,self.dofTypes.index(dofType)] )
            self.constrainedVals.append( val )

    def getForType ( self, nodeIDs, dofType ):
        return self.dofs[self.IDmap.get( nodeIDs ),self.dofTypes.index(dofType)]

    def get ( self, nodeIDs ):
        return self.dofs[self.IDmap.get(nodeIDs)].flatten()

    def getConstraintsMatrix ( self ):

        n_constrained = len( self.constrainedDofs )
        n             = len( self )

        C = np.zeros( (n,n-n_constrained) )
        j = 0
        
        for i in range(n):
        
            if i in self.constrainedDofs:
                continue

            C[i,j] = 1.
            j+=1

        return C

    def solve ( self, A, b ):

        if len(A.shape) == 2:
            C = self.getConstraintsMatrix()

            a = np.ALLOW_THREADSzeros(len(self))
            a[self.constrainedDofs] = self.constrainedFac * array(self.constrainedVals)

            A_constrained = np.dot( dot( C.transpose(), A ), C )
            b_constrained = np.dot( C.transpose(), b + np.dot( A , -a ) )

            x_constrained = scipy.linalg.solve( A_constrained, b_constrained )

            x = np.dot( C, x_constrained )

            x[self.constrainedDofs] = self.constrainedFac * np.array(self.constrainedVals)
            
        elif len(A.shape) == 1:
            x = b / A

            x[self.constrainedDofs] = self.constrainedFac * array(self.constrainedVals)
        
        return x
    
    def eigensolve( self, A , B ):

        C = self.getConstraintsMatrix()

        A_constrained = ( C.T @ A ) @ C 
        B_constrained = ( C.T @ B ) @ C 

        #x_constrained = scipy.linalg.solve( A_constrained, b_constrained )

        return A,B

    def norm ( self, r ):

        C = self.getConstraintsMatrix()
        
        return scipy.linalg.norm( C.T @ r )

##>*****************************************************************************

##<****************************Abstract objects*********************************
class Material():
    def __init__(self, tag, E, nu=None):
        self.tag = tag
        self.E: float = E
        self.elastic_modulus: float = E 
        self.poisson_ratio = nu

class XSect():
    def __init__(self, tag, A, I):
        self.tag = tag
        self.A: float = A
        self.I: float = I
    
def  UnitProperties(): # Legacy; consider removing and adding unit materials to model by default
    return (Material('unit', 1.0), XSect('unit', 1.0, 1.0))

class State():
    """STATE is a data structure with information about the current state of the structure in fields
    
    """
    def __init__(self, model, method="Linear"):
        self.model = model
        self.num = len(model.states)
        self.data = {"Q": [[qi for qi in elem.q.values()] for elem in model.elems],
                     "P": {str(dof):0 for dof in [item for sublist in model.DOF for item in sublist]},
                     "DOF": 'model.numDOF(model)'
                    }
        self.method = method


    def eload(self, elem, mag, dirn='y'):
        if type(elem) is str:
            elem = self.model.delems[elem]    
        if not(type(mag) is list):
            if dirn=='y':
                mag = [0.0, mag]
            elif dirn=='x':
                mag = [mag, 0.0]
        elem.w[self.num] = mag

##>*****************************************************************************

##<****************************Model objects************************************
# Model objects/classes
# These should be created using the methods above
#******************************************************************************
class Node():
    def __init__(self, model, tag: str, ndf, xyz, mass=None):
        if mass is None: mass=0.0

        self.xyz = np.array([xi for xi in xyz if xi is not None])

        self.tag = tag
        self.xyz0 = self.xyz # coordinates in base configuration (unstrained, not necessarily unstressed).
        self.xyzi = self.xyz # coordinates in reference configuration.  

        self.x0: float = xyz[0] # x-coordinate in base configuration (unstrained, not necessarily unstressed).  
        self.y0: float = xyz[1] # y-coordinate in base configuration (unstrained, not necessarily unstressed).  
        self.z0: float = xyz[2] # z-coordinate in base configuration (unstrained, not necessarily unstressed).  

        # Attributes for nonlinear analysis
        # self.xi: float = x # x-coordinate in reference configuration.  
        # self.yi: float = y # y-coordinate in reference configuration.  
        # self.zi: float = z # z-coordinate in reference configuration.  
        
        self.x: float = xyz[0]
        self.y: float = xyz[1]
        self.z: float = xyz[2]

        
        self.rxns = [0]*ndf
        self.model = model
        self.mass = mass
        self.elems = []

        self.p = {dof:0.0 for dof in model.ddof}
        
    def __repr__(self):
        return 'nd-{}'.format(self.tag)

    def p_vector(self):
        return np.array(list(self.p.values()))

    # def p_tensor(self):
    #     p = tf.Variable([[p] for p in self.p.values()])
    #     return p
        
    @property
    def dofs(self):
        # if self.model.DOF == None: self.model.numDOF()
        idx = self.model.nodes.index(self)
        return np.array(self.model.DOF[idx],dtype=int)


class Rxn():
    def __init__(self, node, dirn):
        self.node = node
        self.dirn = dirn

    def __repr__(self):
        return 'rxn-{}'.format(self.dirn)


class Hinge():
    def __init__(self, elem, node):
        self.elem = elem
        self.node = node

# ##>*****************************************************************************

def parabolicArch(length, height):
    pass

def spacetruss(ns, Ro, Ri, H):
    """Macro for generating 3D space truss"""

    alpha = { i -96:chr(i) for i in range(ord("a"), ord("a") + 26)}
    model = Model(3,3)
    m1 = model.material('default', 1.0)
    s1 = model.xsection('default', 1.0, 1.0)
    
    # specify node coordinates for support points
    # angle for supports
    phi = np.arange(ns)/ns*2*np.pi

    # coordinates for support points
    X  = np.cos(phi)*Ro
    Y  = np.sin(phi)*Ro
    # generate support points with height Z of 0
    for i in range(ns):
        model.node(str(i+1), X[i], Y[i], 0.0)
    
    # angles for upper ring (offset by pi/ns degrees from supports)
    phi = phi+np.pi/ns

    # coordinates for upper ring
    X   = np.append(X, np.cos(phi)*Ri)
    Y   = np.append(Y, np.sin(phi)*Ri)

    # generate coordinates for upper ring with height H
    for i in np.arange(ns, 2*ns):
        model.node(str(i+1), X[i], Y[i], H)
        
    for i, j, k in zip(np.arange(ns), np.arange(0, ns), np.arange(ns, 2*ns)):
        model.truss3d(alpha[i+1], model.nodes[j], model.nodes[k], m1, s1)
        
    model.truss3d(alpha[ns+1], model.nodes[0], model.nodes[2*ns-1], m1, s1)
    
    for i, j, k in zip(np.arange(ns+1, 2*ns), np.arange(1, ns), np.arange(ns, 2*ns-1)):
        model.truss3d(alpha[i+1], model.nodes[j], model.nodes[k], m1, s1)

    for i, j, k in zip(np.arange(2*ns, 3*ns-1), np.arange(ns, 2*ns-1), np.arange(ns+1, 2*ns)):
        model.truss3d(alpha[i+1], model.nodes[j], model.nodes[k], m1, s1)

    model.truss3d(alpha[3*ns], model.nodes[ns], model.nodes[2*ns-1], m1, s1)
    
    
    # boundary conditions
    for node in model.nodes[0:ns]:
        model.fix(node, ['x', 'y', 'z'])
        
    model.numDOF()

    return model

class Domain(Model):
    """Deprecated. Use Model instead."""
    pass



import warnings 
from collections import defaultdict
from functools import partial

import numpy as onp
try: 
    import jax
    import jax.numpy as np
except: 
    np=onp



def integrate(f, gradf, hessf=None, ff=None, nr=None,verbose=False, 
            stepper=None, soptions={}, iterate=None, itoptions={},
            init_solve=None, inoptions={}):

    ##########################################################################
    ## initialize strategy
    ##########################################################################

    if nr is not None: nr = -nr

    ## State Handler
    if callable(init_solve):
        pass 
    else:
        inoptions.setdefault('verbose',verbose)
        inoptions.setdefault('nr',nr)
        init_solve = _init_solve(f,gradf,**inoptions)

    ## Iterator 
    if callable(iterate): 
        pass
    else:
        itoptions.setdefault('nr',nr)
        itoptions.setdefault('verbose',verbose)
        iterate = iterate_init(f, gradf, hessf, **itoptions)

    ## Stepper
    if callable(stepper): pass
    else:
        if not soptions: # soptions is empty
            pass
        stepper = lambda state0, **opts:  [state0]
    

    ##########################################################################
    ## Solver
    ##########################################################################
    def solve(x0, params={}, ff=None):
        a = params
        x, a, state0 = init_solve(x0, a, ff)

        for step,state in enumerate(stepper(state0)):

            if verbose: print( 'step: ', step + 1)
            
            x, _ = iterate(x, a, state)

        return x

    return solve



def _init_solve(f, gradf, verbose=False,nr=None,**kwargs):

    def init_solve(x0, a={}, ff=None):
        if verbose: print('Initializing iterative solve.')
        state = {}
        x = x0

        f0 = f(x0, **a)[:nr]
        state['fi'] = f0
        if verbose: print('\nf0: ',np.around(f0,3))

        if ff is None: ff = np.zeros(f0.shape)
        elif callable(ff): ff = ff(x0)[:nr]
        if verbose: print('\nff: ',np.around(ff,3))

        state.setdefault('df',  ff-f0)
        state.setdefault('ff',     ff)
        state.setdefault('nr',     nr)
        return x, a, state
    
    return init_solve

# @jax.custom_jvp
def iterate_init(f, gradf, hessf=None, nr=None, maxiter=20,
                 tol=1e-3,loss=None,verbose=False,jit=True,**kwargs):
    if loss is None:
        loss = np.linalg.norm
    
    def update(x, a, state):
        if verbose: print('x: ', x)
        df = state['df']
        if verbose: print('df: ',df.T)

        gradfi = gradf(x,**a)[:nr,:nr]
        if verbose: print('gradf: \n',gradfi)

        dx = np.linalg.solve(gradfi, df)[:,0]

        if nr is not None:
            dx = np.pad(dx, (0,-nr), 'constant') 
        if verbose: print('dx: ',dx.T)
        x = x + dx

        state['fi'] = f(x, **a)[:nr]
        # state['df'] = df
        return x, state

    if jit:
        try:
            update = jax.jit(update)
            print('Iteration updater jit compilation successful.')
        except:
            pass
    
    def iterate(x0, a, state):
        x = x0
        for itr in range(maxiter):
            if verbose: print('iteration: ',itr)
            state['df'] = state['ff'] - state['fi']
            
            if loss(state['df']) <= tol: 
                return x, state

            x, state = update(x, a, state)

        if verbose:
            msg = ("Failed to converge after %d iterations, value is %s."
                % (itr + 1, loss(state['df'])))
            raise RuntimeError(msg)
        

    return iterate



from scipy.optimize import fsolve 
import numpy as np

def return_map(E,H0,C,AD=False,H=None):
    if H is None:
        H = lambda Δep: H0

    def g(Δep, sigma_trial, q_i, Y_i):
        return abs(sigma_trial - q_i) - (E + C)*Δep - Y_i - H( sigma_trial - (E + C)*Δep )*Δep
    
    if AD:
        Δg = jax.grad(g)
    else:
        Δg = None
        
    def solve_g(sigma_trial, q_i, Y_i):
        return fsolve(g, 0, args=(sigma_trial, q_i, Y_i))[0]
 
    def step(Δe, e_i, ep_i, sigma_i, q_i, Y_i):  # Δe, e_i, sigma_i, q_i, Y_i, ep_i 
        e_n = e_i + Δe                           # New total strain at the end of this step
        sigma_trial = sigma_i + E * Δe           # Trial stress
        xi_trial = sigma_trial  - q_i
        f_trial = abs(xi_trial) - Y_i

        if f_trial < 0: # Elastic step
            sigma_n = sigma_trial
            Y_n = Y_i
            ep_n = ep_i
            q_n = q_i
            
        else: 
            Δep = solve_g(sigma_trial, q_i, Y_i)

            # Update ep_n, sigma_n, q_n, Y_n
            ep_n = ep_i + Δep * np.sign(xi_trial)
            sigma_n = E*(e_n - ep_n)
            Y_n =  Y_i + H(Y_i) * Δep

            q_n = q_i + C*Δep*np.sign(xi_trial)

        return e_n, ep_n, sigma_n, q_n, Y_n 
    return step


import numpy as np 

from emma.objects import Material


class Material1D(Material):
    pass 

class Material2D:
    def __init__(self,E,nu,G=None, eps0=None, sig0=None, Case=None, irs=None):
        self.elastic_modulus = self.E = E

        if G is None:
            self.shear_modulus = self.G = E/2/(nu+1)
        else:
            self.shear_modulus = self.G = G
        if nu is None:    self.nu = 0.0
        if eps0 is None:  self.eps0 = np.zeros((4,1))
        if sig0 is None:  self.sig0 = np.zeros((4,1))
        if Case is None:  self.Case = 'stress'
        if ir is None:
            self.irs = np.array([1, 2, 4])
            self.ics = 3
        else:
            self.ics = setdiff(1:4,MatData.irs)
            self.eps0 = eps0 # initial strain tensor in 4x1 array form in the order 11, 22, 33, 12
            self.sig0 = sig0 # initial stress tensor in 4x1 array form in the order 11, 22, 33, 12
            self.Case = Case # 'stress' or 'strain'
            self.irs  = irs  # stress or strain components to be retained depending on Case
        pass 

class AxisyMat(Material2D):
    pass 

class PlaneStrainMat(Material2D):
    pass 

class PlaneStressMat(Material2D):
    pass

class GMP:
    def __init__(self,fy,Eh,E,r):
        
        def sig(eps):
            xi = eps/(fy/E)
            b = Eh/E
            return fy*(xi*b+(1-b)*xi/(1+abs(xi)**r)**(1/r))

        def Et(eps):
            xi = eps/(fy/E)
            b = Eh/E
            return E*(b+(1-b)/(1+abs(xi)**r)**(1+1/r))
        
        return sig, Et



import tensorflow as tf 

class ElemGenerator:
    def __init__(self, ElemData: dict):
        self.ElemData = ElemData

    def generate_ElemData(self):
        # generates a dictionary between a person and all the photos of that person
        ElemData = dict(self.ElemData)
        return ElemData





"""
Element Library. (:mod:`elemlib`)
=====================================================

.. currentmodule:: matrices

This module provides a collection of various elements.

"""

class LE2dFrm:
    """LE2dFRM 2d LE frame element under linear or nonlinear geometry
    response of 2d linear elastic frame element;
    the element accounts for linear and nonlinear geometry for the nodal dof transformations; 
    depending on the value of the character variable ACTION the function returns information
    in data structure ELEMRESP for the element with number EL_NO, end node coordinates XYZ,
    and material and loading properties in the data structure ELEMDATA."""
    def __init__(self):
        pass 

    def ke_matrix(self,u):
        """Stiffness matrix"""
        pass
    
    def kg_matrix(self,u):
        """Stiffness matrix"""
        pass

    def ag_matrix(self):
        pass



__version__ = "0.0.0"

# __all__ = ['a', 'b', 'c']


import emma.analysis

from emma.objects import *
from emma.matrices import *

from emma.graphics import *
from emma.solvers import *

import emma.utilities
import emma.matlib


settings = {
    "DATAFRAME_LATEX": True,
}





import numpy as np 
import scipy.linalg
import pandas as pd

import emma.matrices as em
# import tensorflow as tf

class LinearGraph:
    """Solve elements in a Model object for free DOF displacements
    
    Parameters
    -------------
    Model: emma.Model
        Model object with predefined geometry and element/nodal loads."""

    def __init__(self,model,grad_vars=None):
        self.model = model 
        self.grad_vars = grad_vars
    
    def run(self):
        mdl = self.model 
        nf = mdl.nf+1
        self.Pf = em.P_tensor(mdl)[0:nf]

        # Form structure stiffness matrix
        self.K = em.K_tensor(mdl)
        self.Kf = self.K[0:nf,0:nf]
        
        Uf = tf.linalg.solve(self.Kf,-self.Pf)

        # Find global free displacements
        self.Uf = Uf
        return Uf

    def gradient(self,wrt):
        mdl = self.model 
        nf = mdl.nf+1
        self.Pf = Pf = em.P_tensor(mdl)[0:nf]
        with tf.GradientTape() as tape:
            tape.watch(wrt)

            # Form structure stiffness matrix
            self.K = em.K_tensor(mdl)
            self.Kf = self.K[0:nf,0:nf]
            
            Uf = tf.linalg.solve(self.Kf,-self.Pf)

            # Find global free displacements
            self.Uf = Uf
        self.grad = tape.gradient(Uf,Pf)
        # self.jacobian = tape.jacobian(Uf,Pf)
        return Uf

class Event2Event():
    def __init__ (self, Model, Pref=None):
        if Pref is None:
            self.Pref = em.P_vector(Model) 
        else:
            self.Pref = Pref
        
        self.Model = Model
        self.init()

    def init(self):
        Model = self.Model
        self.k =  0
        self.Af = em.A_matrix(Model).f
        self.nf = Model.nf
        # set up element diagonal flexibility matrix, Fs
        self.Fs  = em.Fs_matrix(Model, Roption=False)
        # extract plastic capacity data from to set up Qpl array
        self.Qpl = em.Qpl_vector(Model)
        # initializations
        self.lamdah   = np.zeros(1)    # load factor
        self.nq       = len(Model.basic_forces)  # number of basic forces in model
        self.Qh  = np.zeros((self.nq,1))    # basic element forces
        self.Ufh = np.zeros((self.nf,1))    # displacements at free dofs
        self.Vph = np.zeros((self.nq,1))    # plastic deformations
        # initialize indices of basic forces, plastic hinge locations and non-trivial equilibrium equations
        self.ind_e = Model.idx_c        # indices of continuous/elastic basic forces
        self.ind_n = np.arange(self.nf)      # indices of non-trivial equilibrium equations     
        self.Iph   = np.array([])       # indices of plastic hinge locations
        # self.Af0 = self.Af[self.ind_e, :].del_zeros()
        self.Af0 = self.Af[self.ind_e, :]


        # form kinematic matrix A for all dofs of structural model and extract Af matrix from free dofs
    def next(self):
        """Increment to next event"""
        
        if np.linalg.matrix_rank(self.Af0)==len(self.ind_n):
            # set up stiffness matrix and solve for free dof displacements
    #         print(Af0)
            self.KsAf = (self.Fs[self.ind_e,:][:,self.ind_e]).inv @ self.Af0
            self.Kf   = self.Af0.T @ self.KsAf

            self.Ufe  = np.zeros((self.nf,1))
            self.Ufe[self.ind_n,0]  = self.Kf.inv @ self.Pref[self.ind_n]
            
            # determine basic element forces Qe (elastic increment)
            # self.Qe = np.zeros((self.nq,1))
            self.Qe = em.Q_vector(self.Model)
            self.Qe[self.ind_e,:] = self.KsAf @ self.Ufe[self.ind_n]
            
            # distinguish positive and negative basic force values
            self.ineg = np.where(self.Qe[self.ind_e,:]<0)[0]
            self.ne   = len(self.ind_e)
            self.ipos = np.where(self.Qe[self.ind_e,:]>=0.0)[0]

            self.ipe  = self.ind_e[self.ipos]
            self.ine  = self.ind_e[self.ineg]

            # set up residual plastic capacity Qplres for elastic locations
            self.Qplres = np.zeros((self.ne,1))
            self.Qplres[self.ipos,0] =  self.Qpl[self.ipe, 0] - self.Qh[self.ipe, self.k]    
            self.Qplres[self.ineg,0] = -self.Qpl[self.ine, 1] - self.Qh[self.ine, self.k] 
            
            # form demand capacity ratio DC for elastic basic forces
            self.DC        = np.zeros((self.nq,1))
            self.DC[self.ind_e] = self.Qe[self.ind_e]/self.Qplres

            
            # determine location of maximum DC
            self.new_hinge = np.where(self.DC == np.amax(self.DC))[0]
            
            # determine load factor increment for next hinge formation (use one hinge only)
            self.Dlamda     = 1/self.DC[self.new_hinge[0]]
            
            # update load factor
            self.lamdah = np.append(self.lamdah, self.lamdah[self.k] + self.Dlamda)

            # update global dof displacements and basic element forces with elastic increment
            self.Ufh = np.append(self.Ufh, np.array([self.Ufh[:,self.k] + self.Ufe@self.Dlamda]).T, axis=1)
            self.Qh = np.append(self.Qh, np.array([self.Qh[:,self.k] + self.Qe@self.Dlamda]).T, axis=1)


            # update list of plastic hinge locations (several hinges can form simultaneously)
            self.Iph   = np.append(self.Iph, self.new_hinge.T)

            # locate trivial equilibrium equations and set displacement equal to previous step
            self.ind_t = np.setdiff1d(list(range(self.nf)), self.ind_n)
            self.Ufh[self.ind_t, self.k+1] = self.Ufh[self.ind_t,self.k]

            # determine plastic deformations by Vpl = Af*Uf - Fs*Q
            self.Vph = np.append(self.Vph, np.array([self.Af@self.Ufh[:,self.k+1] - self.Fs@self.Qh[:,self.k+1]]).T, axis=1)

            # update list of elastic basic forces by removing new plastic hinge locations
            self.ind_e = np.setdiff1d(self.ind_e, self.new_hinge.T)

            # remove any columns of Af with all zero entries (trivial equilibrium equations)
            
            self.Af0 = self.Af[self.ind_e, :]
            # increment step counter
            self.k += 1
            
        return self.lamdah, self.Qh
        
        
    def run(self):  
        Pref = self.Pref
        Model = self.Model 
            # form kinematic matrix A for all dofs of structural model and extract Af matrix from free dofs
        Af = em.A_matrix(Model).f

        nf = Model.nf

        # set up element diagonal flexibility matrix, Fs
        Fs  = em.Fs_matrix(Model)

        # extract plastic capacity data from to set up Qpl array
        Qpl = em.Qpl_vector(Model)

        # initializations
        k        = 0                        # load step counter
        lamdah   = np.zeros(1)              # load factor
        nq       = len(Model.basic_forces)  # number of basic forces in model
        Qh  = np.zeros((nq,1))              # basic element forces
        Ufh = np.zeros((nf,1))              # displacements at free dofs
        Vph = np.zeros((nq,1))              # plastic deformations
        Acp = np.zeros((nq,1))              # plastic deformations
        Amp = np.zeros((nf,1))
        # initialize indices of basic forces, plastic hinge locations and non-trivial equilibrium equations
        ind_e = Model.idx_c        # indices of continuous/elastic basic forces
        ind_n = np.arange(nf)      # indices of non-trivial equilibrium equations     
        Iph   = np.array([])       # indices of plastic hinge locations
        
        # Loop until the structure is unstable
        # Af0 = Af[ind_e, :].del_zeros()
        Af0 = Af[ind_e, :]

        while np.linalg.matrix_rank(Af0)==len(ind_n):
            # set up stiffness matrix and solve for free dof displacements

            KsAf = (Fs[ind_e,:][:,ind_e]).inv @ Af0
            Kf   = Af0.T @ KsAf

            Ufe  = np.zeros((nf,1))
            Ufe[ind_n,0]  = Kf.inv @ Pref[ind_n]
            
            # determine basic element forces Qe (elastic increment)
            # Qe = np.zeros((nq,1))
            Qe = em.Q_vector(Model)
            Qe[ind_e,:] = KsAf @ Ufe[ind_n]
            
            # distinguish positive and negative basic force values
            ineg = np.where(Qe[ind_e,:]<0)[0]
            ne   = len(ind_e)
            ipos = np.where(Qe[ind_e,:]>=0.0)[0]

            ipe  = ind_e[ipos]
            ine  = ind_e[ineg]

            # set up residual plastic capacity Qplres for elastic locations
            Qplres = np.zeros((ne,1))
            Qplres[ipos,0] =  Qpl[ipe, 0] - Qh[ipe, k]    
            Qplres[ineg,0] = -Qpl[ine, 1] - Qh[ine, k] 
            
            # form demand capacity ratio DC for elastic basic forces
            DC        = np.zeros((nq,1))
            DC[ind_e] = Qe[ind_e]/Qplres

            
            # determine location of maximum DC
            new_hinge = np.where(DC == np.amax(DC))[0]
            
            # determine load factor increment for next hinge formation (use one hinge only)
            Dlamda     = 1/DC[new_hinge[0]]
            
            # update load factor
            lamdah = np.append(lamdah, lamdah[k] + Dlamda)

            # update global dof displacements and basic element forces with elastic increment
            Ufh = np.append(Ufh, np.array([Ufh[:,k] + Ufe@Dlamda]).T, axis=1)
            Qh = np.append(Qh, np.array([Qh[:,k] + Qe@Dlamda]).T, axis=1)


            # update list of plastic hinge locations (several hinges can form simultaneously)
            Iph   = np.append(Iph, new_hinge.T)

            # locate trivial equilibrium equations and set displacement equal to previous step
            ind_t = np.setdiff1d(list(range(nf)), ind_n)
            Ufh[ind_t, k+1] = Ufh[ind_t,k]

            # determine plastic deformations by Vpl = Af*Uf - Fs*Q
            Vph = np.append(Vph, np.array([Af@Ufh[:,k+1] - Fs@Qh[:,k+1]]).T, axis=1)
            
            # Ampi = scipy.linalg.null_space(Af[ind_e, :])
            # Amp = np.append(Amp, Ampi, axis=1)
            # Acp = np.append(Acp,  Af @ Ampi, axis=0)
            

            # update list of elastic basic forces by removing new plastic hinge locations
            ind_e = np.setdiff1d(ind_e, new_hinge.T)

            # remove any columns of Af with all zero entries (trivial equilibrium equations)
            
            Af0 = Af[ind_e, :]
            # increment step counter
            k += 1
        
        output = (lamdah, Qh, Ufh, Vph, Iph)
        self.lamda = lamdah 
        self.Q  = [em.Q_vector(self.Model, Qh[:,K].T) for K in range(k+1)]
        # self.U = [em.U_vector(self.Model, Ufh[:,K].T) for K in range(k+1)]  
        self.U = pd.DataFrame(Ufh)
        self.V = [em.V_vector(self.Model, Vph[:,K].T) for K in range(k+1)]  
        self.Iph = Iph
        self.Acp = Acp 
        self.Amp = Amp 
        return output[0:2]


"""
Matrices. (:mod:`matrices`)
=====================================================

.. currentmodule:: matrices

This module provides functions and classes for constructing various
structural analysis matrices. Matrices are generally constructed with
functions that are defined outside of classes (as opposed to using the
__init__ method of a class) in an effort to present basic structural analysis
concepts with minimal clutter from computer science concepts which may
be unfamiliar to engineers.

"""

import numpy as np
import pandas as pd
import scipy.linalg
import matplotlib.pyplot as plt
# import sympy as sp
# import tensorflow as tf
import emma
# from emma.utilities import Structural_Matrix, Structural_Vector, transfer_vars


settings = {
    "DATAFRAME_LATEX": True,
}

subscripts = {
    'part': 'subscript',
     None: '',
    'initial': '0',
    'continuous':'c',
    'primary':'i',
    'reactions':'d',
    'elem-load':'w',
    'free': 'f',
}

def del_zeros(mat):
    delrows = np.where(~mat.any(axis=1))[0]
    delcols = np.where(~mat.any(axis=0))[0]

    newM = np.delete(mat, delcols, axis=1)
    newM = np.delete(newM, delrows, axis=0)
    return newM

def elem_dofs(Elem):
    dofs = []
    for node in Elem.nodes:
        dofs.extend(node.dofs)
    return dofs

def transfer_vars(item1, item2):
    for key, value in item1.__dict__.items():
        item2.__dict__[key] = value

def Localize(U_vector, P_vector, model=None):
    if model is None: model = U_vector.model
    A =  A_matrix(model)
    Q0 = Q0_vector(model)
    Ks = Ks_matrix(model)

    V = A.f @ U_vector.f
    Q  = Ks@V + Q0
    return V, Q



class Structural_Vector(np.ndarray):
    column_data = ["vector"]
    row_data = None
    subs = [None]
    tag = 'Vector'
    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def _repr_html_(self):
        try:
            out = self.df.to_html()
            return out
        except: 
            return self

    def __add__(self, other):
        if isinstance(other, type(self)):
            out = np.add(self, other).view(type(self))
            transfer_vars(self, out)
        else:
            out = super().__add__(other)
        return out
    
    def get(self, key):
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key], dtype=int)
        # row = self.row_data.index(component)
        return self[idx]

    def set_item(self, key, value):
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key], dtype=int)
        self[idx] = value

    def rows(self, component):
        idx = np.where(np.isin(self.row_data,component))[0]
        newV = self[idx]
        newV.row_data = np.array(self.row_data)[idx]
        newV.model = self.model
        return newV

    @property
    def df(self):
        row_data = ['$'+str(tag)+'$' for tag in self.row_data]
        header = '$'+self.tag+'_{{'
        try:
            for sub in self.subs:
                header += subscripts[sub]
        except: pass
        print
        header += '}}$'

        return pd.DataFrame(np.around(self,14), index=row_data, columns=[header])

    @property
    def symb(self):
        var = []
        for eid in self.row_data:
            var.append(sp.symbols(self.tag+eid))
        return sp.Matrix(var)
    
    @property
    def disp(self):
        return sp.Matrix(self)
     
class Structural_Matrix(np.ndarray):
    column_data = None
    row_data = None
    c_ridx = None # row indexes for use in .c method
    c_cidx = None # column indexes for use in .c method
    tag = None

    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def __matmul__(self, other):
        if isinstance(other, Structural_Matrix):
            out = np.matmul(self,other).view(Structural_Matrix)
            transfer_vars(self,out)
            out.column_data = other.column_data
        
        elif isinstance(other, Structural_Vector):
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            # out.column_data = ['.']
        else:
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            # out.column_data = ['.']
        return out

    def __add__(self, other):
        out = np.add(self,other)
        if (isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out.row_data = self.row_data
        return out

    
    
    def __truediv__(self, other):
        out = np.ndarray.__truediv__(self, other) 
        if (isinstance(other, float) 
            or isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out = np.ndarray.__truediv__(self, other).view(Structural_Matrix)
            transfer_vars(self, out)
        else:
            out = np.ndarray.__truediv__(self, other)
        return out
        
    def _repr_html_(self):
        try: 
            df = self.df
            return df.to_html()
        except:
            try:
                return pd.DataFrame(self).to_html()
            except:
                pass

    @property
    def disp(self):
        return sp.Matrix(self)
    
    @property
    def df(self):
        if settings['DATAFRAME_LATEX']:
            row_data = ['$'+str(tag)+'$' for tag in self.row_data]
            column_data = ['$'+str(tag)+'$' for tag in self.column_data]
        else:
            row_data = [str(i) for i in self.row_data]
            column_data = [str(i) for i in self.column_data]
        return pd.DataFrame(np.around(self,5), index=row_data, columns=column_data)

    @property
    def inv(self):
        mat = np.linalg.inv(self)
        transfer_vars(self, mat)
        mat.row_data = self.column_data
        mat.column_data = self.row_data
        return mat
    
    @property
    def rank(self):
        """Return the rank of a matrix"""
        return np.linalg.matrix_rank(self)
    
    @property
    def lns(self):
        """Return a basis for the left nullspace of a matrix."""
        return scipy.linalg.null_space(self.T)
    
    @property
    def nls(self):
        """return a basis for the nullspace of matrix."""
        return scipy.linalg.null_space(self)
    
    @property
    def ker(self):
        "Return a basis for the kernel (nullspace) of a matrix."
        kernel = scipy.linalg.null_space(self) 
        ker = Structural_Matrix(kernel)
        transfer_vars(self,ker)
        ker.row_data = self.column_data 
        ker.column_data = [str(i+1) for i in range(len(ker[0]))]
        return ker

    def lu(self):
        
        return scipy.linalg.lu(self)

    @property
    def c(self):
        delcols = self.c_cidx
        delrows = self.c_ridx

        newM = np.delete(self, delrows, axis=0).view(type(self))
        if delcols: newM = np.delete(newM, delcols, axis=1).view(type(self))
        
        transfer_vars(self, newM)

        if delcols: newM.column_data = list(np.delete(self.column_data, delcols))

        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM

    def round(self, num):
        newM = np.around(self, num).view(Structural_Matrix)
        transfer_vars(self,newM)
        return newM

    def remove(self, component):
        """Remove items by looking up column_data/row_data"""
        if type(component) is list:
            for item in component:
                if item in self.column_data: 
                    delcol = self.column_data.index(item)
                    newM = np.delete(self, delcol, axis=1).view(type(self))
                    transfer_vars(self,newM)
                    newM.column_data = list(np.delete(self.column_data, delcol))
                else:
                    delrow = self.row_data.index(item)
                    newM = np.delete(self, delrow, axis=0).view(type(self))
                    transfer_vars(self, newM)
                    # newM.column_data = self.column_data
                    # newM.model = self.model
                    newM.row_data = list(np.delete(self.row_data, delrow))

        else:
            item = component
            if item in self.column_data: 
                delcol = self.column_data.index(item)
                newM = np.delete(self, delcol, axis=1).view(type(self))
                newM.row_data = self.row_data
                newM.model = self.model
                newM.column_data = list(np.delete(self.column_data, delcol))
                # try: newM.rel = self.rel
                # except: pass
            else:
                delrow = self.row_data.index(item)
                newM = np.delete(self, delrow, axis=0).view(type(self))
                newM.column_data = self.column_data
                newM.model = self.model
                newM.row_data = list(np.delete(self.row_data, delrow))
                # try: newM.rel = self.rel
                # except: pass
        return newM

    def get(self, row_name, col_name):
        idxr = np.where(self.row_data == row_name)
        idxc = np.where(self.column_data == col_name)
        return self[idxr, idxc]


    def rows(self, component):
        rows = [self.row_data.index(item) for item in component]
        newM = self[rows,:]
        newM.model = self.model
        newM.column_data = self.column_data
        newM.row_data = list(np.array(self.row_data)[rows])
        return newM

    def del_zeros(self):
        """Delete rows and columns of a matrix with all zeros"""
        delrows = np.where(~self.any(axis=1))[0]
        delcols = np.where(~self.any(axis=0))[0]

        newM = np.delete(self, delcols, axis=1).view(type(self))
        newM = np.delete(newM, delrows, axis=0).view(type(self))
        transfer_vars(self, newM)

        newM.column_data = list(np.delete(self.column_data, delcols))
        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM
        
    def add_cols(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other
        if type(component) is list:
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newM = np.delete(self, delcols, axis=1).view(type(self))
            for col in delcols:
                newM[:,i0] += self[:,col]
            newM.row_data = self.row_data
            newM.model = self.model
            newM.column_data = list(np.delete(self.column_data, delcols))
            # try: newM.rel = self.rel
            # except: pass
            return newM

    def add_rows(self, component):
        if type(component) is list:
            delrows = [self.row_data.index(item) for item in component[1:len(component)]]
            i0 = self.row_data.index(component[0])
            newA = np.delete(self, delrows, axis=0).view(type(self))
            for row in delrows:
                newA[i0,:] += self[row,:]
            newA.column_data = self.column_data
            newA.model = self.model
            newA.row_data = list(np.delete(self.row_data, delrows))
            return newA

class row_vector (Structural_Vector):
    def __new__(cls, Matrix):
        V = np.zeros((len(Matrix.row_data)))
        return np.asarray(V).view(cls)

    def __init__(self, Matrix):
        self.tag = "Y"
        self.matrix = Matrix 
        self.row_data = Matrix.row_data
    
class column_vector (Structural_Vector):

    def __new__(cls, Matrix, Vector=None):
        V = np.zeros((len(Matrix.column_data)))
        return np.asarray(V).view(cls)

    def __init__(self, Matrix, Vector=None):
        self.tag = "X"
        self.matrix = Matrix 
        self.row_data = Matrix.column_data
        if Vector is not None:
            for key in Vector.row_data:
                self.set_item(key, Vector.rows([key]))


class Static_matrix (Structural_Matrix):
    """B_MATRIX static matrix of structural model with 2d/3d truss and 2d frame elements
    the function forms the static matrix B for all degrees of freedom and
    all basic forces of the structural model specified in data structure MODEL;
    the function is currently limited to 2d/3d truss and 2d frame elements
    
    Parameters
    ---------------

    model: emma.Model object
    
    Partitions
    =========================================================================================
 
    - B.f  : nf x ntq

    - B.c  : nf x nq

    - B.fc : nf x nq

    - B.i  : ni x nq

    - B.x  : nx x nq

    where:

    - ni: number of primary (non-redundant) forces.
    - nq: number of total, continuous forces.
    - nx: number of redundant forces.

    """

    ranges = {
        'f': 'free dof rows',
        'i': 'primary force columns',
        'x': 'redundant force columns',
        'c': 'continuous force columns (removes releases)',
        'd': 'reaction force columns',
    }

    def __new__(cls, model, matrix=None, rng=None):
        fullnq = sum([len(elem.rel) for elem in model.elems])
        B = np.zeros((model.nt,fullnq))
        ci = 0
        for elem in model.elems:
            dofs = elem.dofs
            bg = elem.bg_matrix()
            nq = np.size(bg,1)
            for j,dof in enumerate(dofs):
                B[int(dof)-1,ci:ci+nq] = bg[j,:]
            ci = ci+nq
        input_array = B
        return np.asarray(input_array).view(cls)

    def __init__(self, model, matrix=None, rng=None):
        if rng is None:
            self.rng = None
        self.model = model
        self.row_data = np.array([str(dof) for dof in range(1, model.nt+1)])
        self.column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        if matrix is not None:
            fullnq = sum([len(elem.rel) for elem in model.elems])
            self[:,:] = np.zeros((model.nt,fullnq))
            for idxr, rw in enumerate(self.row_data):
                if rw in matrix.row_data:
                    for idxc, cl in enumerate(self.column_data):
                        if cl in matrix.column_data:
                            self[idxr,idxc] = matrix.get(rw,cl)

    def __matmul__(self, Vector):
        if type(Vector) is nForce_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is iForce_vector:
            vect = np.matmul(self, Vector).view(nForce_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect
  #<# 
    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newB)
        # newB.model = self.model
        newB.row_data = list(np.delete(self.row_data, delrows))
        # newB.column_data = self.column_data
        return newB

    @property
    def i(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces"""
        Bf = self.f
        idx_i = self.model.idx_i
        newB = Bf[:,idx_i]
        transfer_vars(Bf, newB)
        newB.column_data = Bf.column_data[idx_i]
        return newB
    
    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newB)
        newB.row_data = list(np.delete(self.row_data, delrows))
        return newB

    @property
    def c(self):

        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        return newB

    @property
    def o(self):
        """Remove columns corresponding to element force releases, then delete zeros"""
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        newB = newB.del_zeros()
        return newB

    @property
    def fc(self):
        return self.f.c

    @property
    def x(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces"""
        idx_x = self.model.idx_x
        newB = self[:,idx_x]
        transfer_vars(self, newB)
        newB.column_data = self.column_data[idx_x]
        return newB

    @property
    def barxi(self):
        Bx = self.f.x
        Bbarxi = self.bari @ -Bx
        Bbarxi.column_data = Bx.column_data
        return Bbarxi

    @property
    def barx(self):
        nQ = len(self.column_data)
        nx = len(self.model.redundants)

        Bbarxi = self.barxi

        Bbarx = Structural_Matrix(np.zeros((nQ,nx)))
        transfer_vars(self, Bbarx)
        Bbarx.column_data = Bbarxi.column_data
        Bbarx.row_data = self.column_data

        for idxc, cl in enumerate(Bbarx.column_data):
            for idxr, rw in enumerate(Bbarx.row_data):
                if rw in Bbarxi.row_data:
                    Bbarx[idxr,idxc] = Bbarxi.get(rw,cl)
                elif cl==rw:
                        Bbarx[idxr, idxc] = 1.

        return Bbarx

    @property
    def bari(self):
        return self.i.del_zeros().inv

    @property
    def c0(self):
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = Bf 
        for col in delcols:
            newB[:,col] = [0.]*len(Bf[:,0])
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        return newB

class nStatic_matrix(Structural_Matrix):
    ranges = {
        'f': 'free dof rows',
        'i': 'primary force columns',
        'x': 'redundant force columns',
        'c': 'continuous force columns (removes releases)',
        'd': 'reaction force columns',}
    def __new__(cls, arry, model, rcdata):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcdata):
        self.model = model
        self.row_data = rcdata[0]
        self.column_data = rcdata[1]
        
    def __matmul__(self, Vector):
        if type(Vector) is nForce_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, iForce_vector):
            vect = np.matmul(self, Vector).view(nForce_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newB)
        newB.row_data = np.delete(self.row_data, delrows)
        return newB
        
    @property
    def i(self):
        """Removes rows of B_matrix corresponding to redundant forces"""
        # reducedB = self.f.c.del_zeros()
        reducedB = self.f.c
        rdts = reducedB.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        delcols = [reducedB.column_data.index(tag) for tag in tags]
        newB = np.delete(reducedB, delcols, axis=1).view(Static_matrix)
        transfer_vars(reducedB, newB)
        newB.column_data = list(np.delete(reducedB.column_data, delcols))
        return newB
    
    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newB)
        newB.row_data = list(np.delete(self.row_data, delrows))
        return newB

    @property
    def c(self):
        """Removes columns corresponding to element hinges/releases"""
        Af = self.f
        idx_c = self.model.idx_c
        newA = Af[:,idx_c]
        transfer_vars(Af, newA)
        newA.column_data = Af.column_data[idx_c]
        return newA

    @property
    def o(self):
        """Remove columns corresponding to element force releases, then delete zeros"""
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        newB = newB.del_zeros()
        return newB

    @property
    def fc(self):
        return self.f.c
    
    @property
    def x(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces
        
        """
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        cols = [self.column_data.index(tag) for tag in tags]
        newB = self[:,cols]
        transfer_vars(self, newB)
        newB.column_data = np.array([self.column_data[col] for col in cols])
        # newB.row_data = self.row_data
        # newB.model = self.model
        return newB

    @property
    def barxi(self):
        Bx = self.f.x
        Bbarxi = self.bari @ -Bx
        Bbarxi.column_data = Bx.column_data
        return Bbarxi

    @property
    def barx(self):
        nQ = len(self.column_data)
        nx = len(self.model.redundants)

        Bbarxi = self.barxi

        Bbarx = Structural_Matrix(np.zeros((nQ,nx)))
        transfer_vars(self, Bbarx)
        Bbarx.column_data = Bbarxi.column_data
        Bbarx.row_data = self.column_data
        for idxc, cl in enumerate(Bbarx.column_data):
            for idxr, rw in enumerate(Bbarx.row_data):
                if rw in Bbarxi.row_data:
                    Bbarx[idxr,idxc] = Bbarxi.get(rw,cl)
                elif cl==rw:
                        Bbarx[idxr, idxc] = 1.

        return Bbarx

    @property
    def bari(self):
        return self.i.del_zeros().inv

    @property
    def c0(self):
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = Bf
        for col in delcols:
            newB[:,col] = [0.]*len(Bf[:,0])
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        # newB.row_data = self.row_data
        # newB.rel = self.rel
        return newB

def B_matrix(model, matrix=None, rng=None):
    """Returns a Static_matrix object"""
    return Static_matrix(model, matrix, rng)

def nB_matrix(model, rng=None):
    """Returns a Static_matrix object"""
    fullnq = sum([len(elem.rel) for elem in model.elems])
    B = np.zeros((model.nt, fullnq))
    ci = 0
    for elem in model.elems:
        eid = elem.dofs
        bg = elem.bg_matrix()
        nq = np.size(bg,1)
        for j,eidi in enumerate(eid):
            B[int(eidi)-1,ci:ci+nq] = bg[j,:]
        ci = ci+nq
    input_array = B
    return np.asarray(input_array).view(nStatic_matrix)
    matrix =  np.asarray(input_array)
    # return nStatic_matrix(model, matrix, rng)

def Bh_matrix(model):
    """Returns a Static_matrix object"""
    fullnq = sum([len(elem.rel) for elem in model.elems])
    B = np.zeros((model.nt, fullnq))
    ci = 0
    for elem in model.elems:
        eid = elem.dofs
        bg = elem.bg_matrix(Roption=True)
        nq = np.size(bg,1)
        for j,eidi in enumerate(eid):
            B[int(eidi)-1,ci:ci+nq] = bg[j,:]
        ci = ci+nq
    row_data = np.array([str(dof) for dof in range(1, model.nt+1)])
    column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
    rcdata = (row_data, column_data)
    return nStatic_matrix(B, model, rcdata)

class Kinematic_matrix (Structural_Matrix):
    """Class for the kinematic matrix of a structural model with 2d/3d truss and 2d frame elements
    the function forms the kinematic matrix A for all degrees of freedom and
    all element deformations of the structural model specified in data structure MODEL
    the function is currently limited to 2d/3d truss and 2d frame elements
    
    Returns
    ---------

    Kinematic matrix

    """

    ranges = {
        'f': 'free dof columns',
        'i': 'primary force/deformation rows',
        'x': 'redundant force/deformation rows',
        'd': 'reaction force/deformation rows',
        'c': 'continuous (hingeless) force/deformation rows'
    }

    def __new__(cls, model, matrix=None,rng=None):
        A  = np.zeros((sum(model.nv),model.nt))
        ri = 0
        for elem in model.elems:
            eid = elem.dofs
            ag = elem.ag()
            nv = len(elem.v)
            for j, eidi in enumerate(eid):
                A[ri:ri+nv, int(eidi)-1] = ag[:,j]
            ri = ri+nv
        input_array = A
        return np.asarray(input_array).view(cls)

    def __init__(self, model, matrix=None,rng=None):
        if rng is None:
            self.rng = None
        self.model = model
        self.column_data = np.array([str(dof) for dof in range(1,model.nt+1)])
        self.row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()])
        self.basic_deformations = np.array([v for elem in model.elems for v in elem.basic_deformations])
        
        self.idx_h = []
        if matrix is not None:
            fullnq = sum([len(elem.rel) for elem in model.elems])
            self[:,:] = np.zeros((model.nt,fullnq))
            for idxr, rw in enumerate(self.row_data):
                if rw in matrix.row_data:
                    for idxc, cl in enumerate(self.column_data):
                        if cl in matrix.column_data:
                            self[idxr,idxc] = matrix.get(rw,cl)

    def __matmul__(self, Vector):
        if isinstance(Vector, Deformation_vector):
            # print('a')
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, Displacement_vector):
            vect = np.matmul(self, Vector).view(Deformation_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)

        return vect


    def combine(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other

        if type(component) is list:
            ## TO BE DEPRECATED
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
            for col in delcols:
                newA[:,i0] += self[:,col]
            newA.row_data = self.row_data
            newA.model = self.model
            newA.column_data = list(np.delete(self.column_data, delcols))
            # newA.rel = self.rel
            return newA
    
    @property
    def f(self):
        """Removes columns corresponding to fixed dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) > self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self, newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA

    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        Afc = self.f.c
        rdts = self.model.redundants
        tags = [q.elem.tag +'_'+ str(q.nature) for q in rdts]
        delrows = [Afc.row_data.index(tag) for tag in tags]
        newA = np.delete(Afc, delrows, axis=0).view(Kinematic_matrix)
        transfer_vars(Afc,newA)
        newA.row_data = list(np.delete(Afc.row_data, delrows))
        return newA

    @property
    def d(self):
        """Removes columns corresponding to free dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) <= self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self,newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA


    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        Af = self.f
        idx_c = self.model.idx_c
        newA = Af[idx_c,:]
        transfer_vars(Af, newA)
        newA.row_data = Af.row_data[idx_c]
        return newA
    
    @property
    def c0(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = np.where(np.isin(Af.row_data,tags))
        # delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def o(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def e(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA


class nKinematic_matrix (Structural_Matrix):
    """Class for the kinematic matrix of a structural model with 2d/3d truss and 2d frame elements
    the function forms the kinematic matrix A for all degrees of freedom and
    all element deformations of the structural model specified in data structure MODEL 
    the function is currently limited to 2d/3d truss and 2d frame elements
    
    Returns
    ---------
    Kinematic matrix

    """

    ranges = {
        'f': 'free dof columns',
        'i': 'primary force/deformation rows',
        'x': 'redundant force/deformation rows',
        'd': 'reaction force/deformation rows',
        'c': 'continuous (hingeless) force/deformation rows'
    }

    def __new__(cls, model, matrix=None,rng=None):
        A  = np.zeros((sum(model.nv),model.nt))
        ri = 0
        for elem in model.elems:
            eid = elem.dofs
            ag = elem.ag()
            nv = len(elem.v)
            for j, eidi in enumerate(eid):
                A[ri:ri+nv, int(eidi)-1] = ag[:,j]
            ri = ri+nv
        input_array = A
        return np.asarray(input_array).view(cls)

    def __new__(cls, arry, model, rcdata):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcdata):
        self.model = model
        self.row_data = rcdata[0]
        self.column_data = rcdata[1]
        self.basic_deformations = np.array([v for elem in model.elems for v in elem.basic_deformations])
        
        self.idx_h = []


    def __matmul__(self, Vector):
        if isinstance(Vector, Deformation_vector):
            # print('a')
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, Displacement_vector):
            # print('b')
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect = np.matmul(self, Vector).view(Structural_Vector)  
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            # print('else')
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    def combine(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other

        if type(component) is list:
            ## TO BE DEPRECATED
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
            for col in delcols:
                newA[:,i0] += self[:,col]
            newA.row_data = self.row_data
            newA.model = self.model
            newA.column_data = list(np.delete(self.column_data, delcols))
            # newA.rel = self.rel
            return newA
    
    @property
    def f(self):
        """Removes columns corresponding to fixed dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) > self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self, newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA

    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        Afc = self.f.c
        rdts = self.model.redundants
        tags = [q.elem.tag +'_'+ str(q.nature) for q in rdts]
        delrows = [Afc.row_data.index(tag) for tag in tags]
        newA = np.delete(Afc, delrows, axis=0).view(Kinematic_matrix)
        transfer_vars(Afc,newA)
        newA.row_data = list(np.delete(Afc.row_data, delrows))
        return newA

    @property
    def d(self):
        """Removes columns corresponding to free dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) <= self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self,newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA


    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        Af = self.f
        idx_c = self.model.idx_c
        newA = Af[idx_c,:]
        transfer_vars(Af, newA)
        newA.row_data = Af.row_data[idx_c]
        return newA
    
    @property
    def c0(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def o(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA
    
    @property
    def e(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        newA = newA.del_zeros()
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA

def A_matrix(Domain, matrix=None):
    """Returns a Kinematic_matrix object"""
    return Kinematic_matrix(Domain,matrix)

class Flexibility_matrix (Structural_Matrix):
    """
    Parameters
    =========================================================================================
    model
    
    -----------------------------------------------------------------------------------------
    """
    tag = 'F'
    def __new__(cls, model, Roption=True):
        f  = np.array([elem.f_matrix() for elem in model.elems])
        Fs = scipy.linalg.block_diag(*f)


        B = Static_matrix(model)
        Fsr = emma.utilities.del_zeros(Fs)

        F = B.bari.T@(Fsr@B.bari)
        
        input_array = np.asarray(F).view(cls)
        input_array.model = model
        input_array.column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        input_array.row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        # input_array.rel = [rel for elem in model.elems for rel in elem.rel.values()]

        # # input_array.xx = np.asarray(Fxx).view(cls)
        # input_array.xx.model = model
        # input_array.xx.row_data = B.x.column_data
        # input_array.xx.column_data = B.x.column_data
        # # input_array.xx.rel = [rel for elem in model.elems for rel in elem.rel.values()]


        input_array.s = np.asarray(Fs).view(cls)
        input_array.s.model = model
        input_array.s.column_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        input_array.s.row_data =    np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
        # input_array.s.rel = [rel for elem in model.elems for rel in elem.rel.values()]

        return input_array

    def __init__(self, model, Roption=True):
        # self.model = model 
        # self.column_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.rel = [rel for elem in model.elems for rel in elem.rel.values()]
        pass

    def __matmul__(self, Vector):
        if type(Vector) is Deformation_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is iForce_vector:
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect.part = 'continuous'
            vect.row_data = self.row_data
            vect.matrix = self

        else:
            vect = Structural_Matrix.__matmul__(self, Vector)

        return vect

    @property 
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        rows = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        newF = np.delete(self, rows, axis=0).view(type(self))
        newF = np.delete(newF, rows, axis=1).view(type(self))
        newF.row_data = list(np.delete(self.row_data, rows))
        newF.column_data = list(np.delete(self.column_data, rows))
        newF.model = self.model
        newF.tag = self.tag + "_c"
        return newF

def F_matrix(Domain):
    """Returns a Flexibility_matrix object"""    
    return Flexibility_matrix(Domain, Roption=True)

class Diag_matrix (Structural_Matrix):
    """Block diagonal matrix of element flexibility/stiffness matrices for structural model
    

    this class represents the block diagonal matrix of element flexibility or stiffness matrices
    for a structural model.
    
    """
    tag = 'F'
    def __new__(cls, arry, rc_data, model):
        basic_forces = rc_data
        arry = np.asarray(arry).view(cls)
        arry.basic_forces = basic_forces
        arry.model = model
        # arry.column_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        arry.column_data = arry.row_data = np.array([q.elem.tag+'_'+str(q.number) for q in basic_forces])
        # arry.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        return arry

    def __init__(self, arry, rc_data, model):
        """Parameters
        =========================================================================================
        model
        
        """
        self.model = model 
        pass

    def __matmul__(self, Vector):
        if type(Vector) is Deformation_vector:
            vect = np.matmul(self, Vector).view(iForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, iForce_vector):
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect.part = 'continuous'
            vect.row_data = self.row_data
            vect.matrix = self

        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property
    def c(self):
        """Removes columns corresponding to element hinges/releases"""
        idx_c = self.model.idx_c
        newA = self[:,idx_c]
        newA = newA[idx_c,:]
        transfer_vars(self, newA)
        newA.basic_forces = self.basic_forces[idx_c]
        newA.column_data = self.column_data[idx_c]
        newA.row_data = self.row_data[idx_c]
        return newA

def Fs_matrix(model, Roption=True):
    """Returns a Flexibility_matrix object"""  
    if Roption:
        f  = np.array([elem.f_matrix(Roption) for elem in model.elems])
        basic_forces = np.array([q for elem in model.elems for q in elem.basic_forces if not q.rel]) 
    else:
        f  = np.array([elem.f_matrix(Roption) for elem in model.elems])
        basic_forces = np.array([q for elem in model.elems for q in elem.basic_forces]) 
    Fs = scipy.linalg.block_diag(*f) 
    
    basic_forces = basic_forces
    Fs = Diag_matrix(Fs, basic_forces, model)
    return Fs

def Ks_matrix(model):
    """Returns a Flexibility_matrix object"""  
    k  = np.array([elem.k_matrix() for elem in model.elems])
    Ks = scipy.linalg.block_diag(*k)  
    basic_forces = np.array([q for elem in model.elems for q in elem.basic_forces])
    Ks = Diag_matrix(Ks, basic_forces, model)
    return Ks


class Stiffness_matrix (Structural_Matrix):
    """...
    Parameters
    =========================================================================================
    model
    
    -----------------------------------------------------------------------------------------
    """
    tag = 'K'
    def __new__(cls, arry, model, Roption=None):

        input_array = np.asarray(arry).view(cls)
        input_array.model = model
        input_array.column_data = [str(dof) for dof in range(1, model.nt+1)]
        input_array.row_data = ['P_{'+str(dof)+'}' for dof in range(1, model.nt+1)]
        
        return input_array

    def __init__(self, arry, model, Roption=None):
        self.subs = [None]
        pass
    

    def __matmul__(self, Vector):
        if type(Vector) is Displacement_vector:
            vect = np.matmul(self, Vector).view(nForce_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is nForce_vector:
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.model = self.model
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK

def K_matrix(Model):
    """Returns a Stiffness_matrix object"""
    K = np.zeros((Model.nt, Model.nt))
    for elem in Model.elems:
        ke = elem.ke_matrix()
        for i, dof in enumerate(elem.dofs):
            for j, doff in enumerate(elem.dofs):
                K[int(dof)-1, int(doff)-1] += ke[i, j]

    return Stiffness_matrix(K, Model, Roption=None)

def K_tensor(Model,U=None):
    """Returns a Stiffness_matrix object"""
    K = np.zeros((Model.nt, Model.nt))
    for elem in Model.elems:
        ke = elem.ke_matrix(U)
        for i, dof in enumerate(elem.dofs):
            for j, doff in enumerate(elem.dofs):
                K[int(dof)-1, int(doff)-1] += ke[i, j]
    return K

def AssemblyTensor(Model):
    pass

def Kt_matrix(Model, State):
    """Returns a Stiffness_matrix object"""
    K = np.zeros((Model.nt, Model.nt))
    for elem in Model.elems:
        kt = elem.kt_matrix(State)
        for i, dof in enumerate(elem.dofs):
            for j, doff in enumerate(elem.dofs):
                K[int(dof)-1, int(doff)-1] += kt[i, j]

    return Stiffness_matrix(K, Model, Roption=None)


class Mass_matrix (Structural_Matrix):
    tag = 'M'
    def __new__(cls, Model):

        M = np.zeros((Model.nt, Model.nt))
        ddof = Model.ddof
        mass_dofs = []
        for node in Model.nodes:
            for i,dof in enumerate(node.dofs):
                if 'r' not in list(ddof.keys())[i]:
                    if node.mass != 0.0: mass_dofs.append(str(dof))
                    M[int(dof)-1, int(dof)-1] += node.mass


        input_array = np.asarray(M).view(cls)
        input_array.model = Model
        input_array.mass_dofs = mass_dofs
        input_array.column_data = ['u_{{'+str(dof)+'}}' for dof in range(1, Model.nt+1)]
        input_array.row_data = ['u_{{'+str(dof)+'}}' for dof in range(1, Model.nt+1)]
        return input_array

    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK
    
    @property
    def m(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK


class Displacement_vector(column_vector):
    tag = 'U'
    def __new__(cls, Kinematic_matrix, Vector=None):
        U = np.zeros((len(Kinematic_matrix.column_data)))
        return np.asarray(U).view(cls)

    def __init__(self, Kinematic_matrix, Vector=None):
        self.matrix = Kinematic_matrix 
        self.row_data = Kinematic_matrix.column_data
        self.subs = [None]
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.rows([key]))

    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        newU = np.delete(self, delrows, axis=0).view(Displacement_vector)
        newU.row_data = list(np.delete(self.row_data, delrows))
        # newU.matrix = self.matrix
        newU.model = self.model
        return newU

class nDisplacement_vector(Structural_Vector):
    tag = 'U'
    def __new__(cls, arry, model, row_data, Vector=None):
        input_array = np.asarray(arry).view(cls)
        return input_array

    def __init__(self, arry, model, row_data, Vector=None):
        self.subs = [None]
        self.model = model
        self.row_data = row_data

        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.rows([key]))

    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        newU = np.delete(self, delrows, axis=0).view(nDisplacement_vector)
        newU.row_data = list(np.delete(self.row_data, delrows))
        newU.model = self.model
        return newU

def U_vector(model, vector=None):
    """Returns a Displacement_vector object"""   
    U = np.zeros(model.nt)
    row_data = [str(dof) for dof in range(1,model.nt+1)]
    U = nDisplacement_vector(U, model, row_data)
    
    if vector is not None:
        if len(vector)==len(U):
            U[:,0] = vector[:]
        else:
            for key in vector.row_data:
                if key in U.row_data:
                    U.set_item(key, vector.rows([key]))
    return U


class iForce_vector(Structural_Vector):
    tag = 'Q'
    def __new__(cls, arry, model, row_data, Vector=None):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.subs = [None]
        self.row_data = row_data
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.get(key))
  #<
    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        rdts = self.model.redundants

        tags = [q.elem.tag + str(q.nature) for q in rdts]
        delrows = [self.row_data.index(tag) for tag in tags]
        newQ = np.delete(self, delrows, axis=0).view(iForce_Vector)
        transfer_vars(self, newQ)
        newQ.subs.append('primary')
        newQ.row_data = list(np.delete(self.row_data, delrows))
        return newQ
    
    @property
    def c(self):
        """Remove rows corresponding to element hinges/releases"""
        idx_c = self.model.idx_c
        newQ = self[idx_c]
        transfer_vars(self, newQ)
        newQ.row_data = self.row_data[idx_c]
        return newQ
    
    @property
    def x(self):
        """Remove rows of corresponding to primary forces"""
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        rows = [self.row_data.index(tag) for tag in tags]
        newV = self[rows]
        newV.row_data = [self.row_data[row] for row in rows]
        transfer_vars(self, newV)
        return newV

def Q_vector(model, vector=None):
    """Returns a iForce_vector object"""   
    
    arry = np.zeros((model.nQ,1))
    row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
    Q = iForce_vector(arry, model, row_data)
    if vector is not None:
        if len(vector)==len(Q):
            Q[:,0] = vector[:]
        else:
            for key in vector.row_data:
                if key in Q.row_data:
                    Q.set_item(key, vector.rows([key]))
    return Q

def Q0_vector(model):
    """Returns a vector of initial element forces"""   
    arry = np.concatenate([elem.q0_vector() for elem in model.elems])
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.q.keys()] 
    return iForce_vector(arry, model, row_data)

def Qpl_vector(model):
    """Returns a vector of element plastic capacities""" 

    Qp_pos = [elem.Qp['+'][key] for elem in model.elems for key in elem.Qp['+']]
    Qp_neg = [elem.Qp['-'][key] for elem in model.elems for key in elem.Qp['-']]
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.Qp['-']]

    # del_idx = np.where(~Bf.any(axis=0))[0]
    # Qp_pos = np.delete(Qp_pos, del_idx)
    # Qp_neg = np.delete(Qp_neg, del_idx)
    # row_data = np.delete(row_data, del_idx)
    column_data = ['Q_{pl}^+', 'Q_{pl}^-']

    Qpl = nKinematic_matrix(np.array([Qp_pos, Qp_neg]).T, model, (row_data, column_data))
    return Qpl

def Qp_vector(model):
    """Returns a vector of element plastic capacities""" 
    B = B_matrix(model)
    Bf = B.f
    Qp_pos = [elem.Qp['+'][key] for elem in model.elems for key in elem.Qp['+']]
    Qp_neg = [elem.Qp['-'][key] for elem in model.elems for key in elem.Qp['-']]
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.Qp['-']]

    del_idx = np.where(~Bf.any(axis=0))[0]
    Qp_pos = np.delete(Qp_pos, del_idx)
    Qp_neg = np.delete(Qp_neg, del_idx)
    row_data = np.delete(row_data, del_idx)
    column_data = ['Q_{pl}^+', 'Q_{pl}^-']

    Qpl = nStatic_matrix(np.array([Qp_pos, Qp_neg]).T, model, (row_data, column_data))
    return Qpl

class nForce_vector(Structural_Vector):
    tag = 'P'
    def __new__(cls, arry, model, row_data, Vector=None):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.subs = [None]
        self.row_data = row_data
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.get(key))


    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newP = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newP)
        # newB.model = self.model
        newP.row_data = list(np.delete(self.row_data, delrows))
        # newB.column_data = self.column_data
        return newP

    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newP = np.delete(self, delrows, axis=0).view(type(self))
        newP.model = self.model
        newP.row_data = list(np.delete(self.row_data, delrows))
        newP.subs.append('reactions')
        return newP
        
  #<

def P_vector(model, vector=None):
    P = np.zeros(model.nt)
    for node in model.nodes:
        p = node.p_vector()
        for i, dof in enumerate(node.dofs):
            P[int(dof)-1] += p[i]
    row_data = [str(dof) for dof in range(1, model.nt+1)]
    return nForce_vector(P, model, row_data)



# def P0_vector(model):
#     """Returns a _ object"""   
#     arry = np.concatenate([elem.p0_vector() for elem in model.elems]) 
#     row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()] 
#     return nForce_vector(arry, model, row_data)

def P0_vector(model):
    P = np.zeros(model.nt)
    for elem in model.elems:
        dofs = elem.dofs
        if hasattr(elem, 'q0_vector'):
            p0 = elem.bg_matrix()@elem.q0_vector()
        else:
            p0 = np.zeros(len(dofs))

        for i,df in enumerate(dofs):
            P[int(df)-1] +=  p0[i]

    row_data = [str(dof) for dof in range(1, model.nt+1)]
    return nForce_vector(P, model, row_data)

def Pw_vector(model):
    P = np.zeros(model.nt)
    for elem in model.elems:
        dofs = elem.dofs
        if len(dofs)==6:
            P[int(dofs[0])-1] +=  elem.w['y']*elem.L/2*elem.sn
            P[int(dofs[1])-1] += -elem.w['y']*elem.L/2*elem.cs
            P[int(dofs[3])-1] +=  elem.w['y']*elem.L/2*elem.sn
            P[int(dofs[4])-1] += -elem.w['y']*elem.L/2*elem.cs
        else:
            pw = elem.pw_vector()
            for i,df in enumerate(dofs):
                P[int(df)-1] +=  pw[i]


    row_data = [str(dof) for dof in range(1, model.nt+1)]
    return nForce_vector(P, model, row_data)


class Deformation_vector(Structural_Vector):
    tag = "V"
    def __new__(cls, arry, model, row_data, Vector=None):
        input_array = np.asarray(arry).view(cls)
        return input_array

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.row_data = row_data
    
    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        idx_c = self.model.idx_c
        newQ = self[idx_c]
        transfer_vars(self, newQ)
        newQ.row_data = self.row_data[idx_c]
        return newQ
    
    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        delrows = [self.row_data.index(tag) for tag in tags]
        newV = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newV)
        newV.row_data = list(np.delete(self.row_data, delrows))
        newV.subs.append('primary')
        return newV

    @property
    def x(self):
        """Removes rows of corresponding to primary forces
        
        """
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        rows = [self.row_data.index(tag) for tag in tags]
        newV = self[rows]
        newV.row_data = [self.row_data[row] for row in rows]
        transfer_vars(self, newV)
        return newV

def V_vector(model, vector=None):
    """Returns a Deformation_vector object"""    
    arry = np.zeros((model.nQ,1))
    row_data = np.array([elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()])
    V = Deformation_vector(arry, model, row_data)
    if vector is not None:
        if len(vector)==len(V):
            V[:,0] = vector[:]
        else:
            for key in vector.row_data:
                if key in V.row_data:
                    V.set_item(key, vector.rows([key]))
    return V

def V0_vector(model):
    """Returns a Deformation_vector object"""   
    arry = np.concatenate([elem.v0_vector() for elem in model.elems]) 
    row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()] 
    return Deformation_vector(arry, model, row_data)

def Aub_matrix(model, alpha):
    """Return the interaction upperbound matrix"""  
    aub  = np.array([elem.aub_matrix(alpha) for elem in model.elems])
    A = scipy.linalg.block_diag(*aub)
    return A



import emma.elements
# from emma.matvecs import P_vector
import numpy as np
from datetime import datetime
FEDEASmap = {
    "2D beam":"Lin2dFrm",
    "2D truss": "LETruss",
}

def ModelData(self):
        ModelData = {
            "nn"        :self.nn,     # property      -number of nodes in structural model
            "ndm"       :self.ndm,    # attribute     -dimension of structural model"
            "XYZ"       :self.XYZ,    #               -node coordinates, nodes are stored columnwise"
            "ne"        :self.ne,     # property      -number of elements"
            "CON"       :self.CON,    # attribute     -node connectivity array"
            "ElemName"  :[elem.type for elem in self.elems],    # property      -cell array of element names"
            "nen"       :[],          #
            "nq(el)"    :[],          # property      -no. of basic forces for element el
            "ndf"       :[],          #
            "nt"        :self.nt,     # property      -total number of degrees of freedom
            "BOUN"      :[],          # attribute     -boundary conditions, nodes are stored columnwise"
            "nf"        :self.nf,     #               -number of free degrees of freedom"
            "DOF"       :self.DOF,    # attribute     -array with degree of freedom numbering, nodes are stored columnwise"
        }
        return ModelData

def FEDEAS(Domain):
    print('CleanStart')

    # Node Definitions
    print('\n% Node Definitions')
    for i, node in enumerate(Domain.nodes):
        print("XYZ({},:) = [{} {}];".format(i+1, node.x, node.y))
    
    # Connectivity
    print('\n% Connections')
    for i, elem in enumerate(Domain.elems):
        ni = Domain.nodes.index(elem.nodes[0])+1
        nj = Domain.nodes.index(elem.nodes[1])+1
        print("CON({},:) = [{} {}];".format(i+1, ni, nj))
    
    # Fixities
    print('\n% Boundary Conditions')
    for i, node in enumerate(Domain.nodes):
        if 1 in node.rxns:
            if Domain.ndf ==2:
                nd = Domain.nodes.index(node)+1
                rx = node.rxns
                print("BOUN({},:) = [{} {}];".format(nd, rx[0], rx[1]))
            if Domain.ndf==3:
                nd = Domain.nodes.index(node)+1
                rx = node.rxns
                print("BOUN({},:) = [{} {} {}];".format(nd, rx[0], rx[1], rx[2]))

    # Element types
    print('\n% Specify element type')
    for i, elem in enumerate(Domain.elems): 
        print("ElemName{{{}}} = '{}';".format(i+1, FEDEASmap[elem.type]))


    print("\n% Create model")
    print("Model = Create_SimpleModel (XYZ,CON,BOUN,ElemName);")

    # Element properties
    print("\n% Element properties")

    print("\n ElemData = cell(Model.ne,1);")
 
    for i, elem in enumerate(Domain.elems): 
        print("\n% Element: {}".format(elem.tag))
        print('ElemData{{{}}}.A = {};'.format(i+1, elem.A))
        print('ElemData{{{}}}.E = {};'.format(i+1, elem.E))
        print('ElemData{{{}}}.Np = {};'.format(i+1, elem.Qpl[0,0]))
        if hasattr(elem,'I'): print('\nElemData{{{}}}.I = {};'.format(i+1, elem.I))

        if elem.nv > 1:
            rel = [1 if rel else 0 for rel in elem.rel.values()]
            print("ElemData{{{}}}.Release = [{};{};{}];".format(i+1, rel[0], rel[1], rel[2]))
            print('ElemData{{{}}}.Mp = {};'.format(i+1, elem.Qpl[1,0]))


    # Element Loads
    print("\n%% Element loads")
    for i, elem in enumerate(Domain.elems): 
        if type(elem) is emma.elements.Beam:
            print("\n% Element: {}".format(elem.tag))
            print('ElemData{{{}}}.w = [{}; {}];'.format(i+1, elem.w['x'], elem.w['y']))

    # Nodal Loads
    print("\n%% Nodal loads")
    print('Pf = zeros({});'.format(Domain.nf))
    for node in Domain.nodes:
        p = node.p_vector()
        for i, dof in enumerate(node.dofs):
            if p[i] != 0.:
                print('Pf({}) = {};'.format(dof, p[i]))
    print('Loading.Pref = Pf;')


class FEDEAS_func:
    def __init__(self,model,filename=None,simple=True):
        if filename is None: filename = 'ReturnModel.m'
        self.time_stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        self.model = model 
        self.filename = filename
        self.simple= simple

    def string(self):
        Domain = self.model
        script = ''
        script += 'function [Model,ElemData,Loading] = {}()'.format(self.filename)


        script += '\n' + 'CleanStart'

        # Node Definitions
        script += '\n' + '\n% Node Definitions'
        for i, node in enumerate(Domain.nodes):
            script += '\n' + "XYZ({},:) = [{} {}];".format(i+1, node.x, node.y)
        
        # Connectivity
        script += '\n' + '\n% Connections'
        for i, elem in enumerate(Domain.elems):
            ni = Domain.nodes.index(elem.nodes[0])+1
            nj = Domain.nodes.index(elem.nodes[1])+1
            script += '\n' + "CON({},:) = [{} {}];".format(i+1, ni, nj)
        
        # Fixities
        script += '\n' + '\n% Boundary Conditions'
        for i, node in enumerate(Domain.nodes):
            if 1 in node.rxns:
                if Domain.ndf ==2:
                    nd = Domain.nodes.index(node)+1
                    rx = node.rxns
                    script += '\n' + "BOUN({},:) = [{} {}];".format(nd, rx[0], rx[1])
                if Domain.ndf==3:
                    nd = Domain.nodes.index(node)+1
                    rx = node.rxns
                    script += '\n' + "BOUN({},:) = [{} {} {}];".format(nd, rx[0], rx[1], rx[2])

        # Element types
        script += '\n' + '\n% Specify element type'
        for i, elem in enumerate(Domain.elems): 
            script += '\n' + "ElemName{{{}}} = '{}';".format(i+1, FEDEASmap[elem.type])


        script += '\n' + "\n% Create model"
        if self.simple:
            script += '\n' + "Model = Create_SimpleModel (XYZ,CON,BOUN,ElemName);"
        else:
            script += '\n' + "Model = Create_Model (XYZ,CON,BOUN,ElemName);"

        # Element properties
        script += '\n' + "\n% Element properties"

        script += '\n' + "\n ElemData = cell(Model.ne,1);"
    
        for i, elem in enumerate(Domain.elems):
            script += '\n' + "\n% Element: {}".format(elem.tag)
            try: script += '\n' + 'ElemData{{{}}}.A = {};'.format(i+1, float(elem.A.numpy()))
            except: script += '\n' + 'ElemData{{{}}}.A = {};'.format(i+1, float(elem.A))

            try: script += '\n' + 'ElemData{{{}}}.E = {};'.format(i+1, float(elem.E.numpy()))
            except: script += '\n' + 'ElemData{{{}}}.E = {};'.format(i+1, float(elem.E))

            script += '\n' + 'ElemData{{{}}}.Np = {};'.format(i+1, elem.Qpl[0,0])

            
            script += '\n' + "ElemData{{{}}}.Geom = 'GL';".format(i+1, elem.Qpl[0,0])
            
            if hasattr(elem,'I'): 
                try: script += '\n' + '\nElemData{{{}}}.I = {};'.format(i+1, float(elem.I.numpy()))
                except: script += '\nElemData{{{}}}.I = {};'.format(i+1, elem.I)

            if elem.nv > 1:
                script += '\n' + 'ElemData{{{}}}.Mp = {};'.format(i+1, elem.Qpl[1,0])
                rel = [1 if rel else 0 for rel in elem.rel.values()]
                script += '\n' + "ElemData{{{}}}.Release = [{};{};{}];".format(i+1, rel[0], rel[1], rel[2])


        # Element Loads
        script += '\n' + "\n%% Element loads"
        for i, elem in enumerate(Domain.elems):
            if type(elem) is emma.elements.Beam:
                script += '\n' + "\n% Element: {}".format(elem.tag)
                script += '\n' + 'ElemData{{{}}}.w = [{}; {}];'.format(i+1, elem.w['x'], elem.w['y'])

        # Nodal Loads
        script += '\n' + "\n%% Nodal loads"
        script += '\n' + 'Pf = zeros({});'.format(Domain.nf)
        for node in Domain.nodes:
            p = node.p_vector()
            for i, dof in enumerate(node.dofs):
                if p[i] != 0.:
                    script += '\n' + 'Pf({}) = {};'.format(dof, p[i])
        script += '\n' + 'Loading.Pref = Pf;'
        return script

    def write(self):
        f = open(self.filename,"w+")
        f.write(self.string())
        f.close()



import matplotlib.pyplot as plt
import numpy as np
from emma.elements import Truss
import emma.matrices as mv
from emma.matrices import U_vector


node_style = [{'color':'black', 'marker':'s', 'markersize':3, 'zorder':2},
                               {'marker':'s', 'markersize':3, 'zorder':2}]

rxn_style = {
    0: {'marker':'>', 'color':'black', 'markeredgewidth':1, 'markersize':9,
            'fillstyle':'none',  'zorder':3,},
    1: {'marker':'^', 'color':'black', 'markeredgewidth':1, 'markersize':9,
            'fillstyle':'none', 'zorder':3,},
    2: {'marker':'s', 'color':'black', 'markeredgewidth':1, 'markersize':9,
            'fillstyle':'none', 'zorder':3,},
    'pin':   {'color':'black', 'marker':'t', 'markersize':5, 'zorder':2},
    'roller':{'color':'black', 'marker':'s', 'markersize':5, 'zorder':2},
    'fixed': {'color':'black', 'marker':'s', 'markersize':5, 'zorder':2}
}

hinge_style = [{'s':20, 'zorder':3, 'facecolors':'white', 'edgecolors':'black'},
               {'s':20, 'zorder':3, 'facecolors':'red', 'edgecolors':'black'}] 

elem_style = [{'color': 'grey', 'linewidth':2, 'linestyle':'-', 'zorder':1},
              {'color':'black', 'linewidth':1, 'linestyle':'-', 'zorder':1}] 

chord_style = [{'color':'black', 'linewidth':1, 'linestyle':':', 'zorder':2}] 

axis_style = {'color':'grey', 'linewidth':1, 'linestyle':'--'}

def plot_moments(state:object, ax=None, scale:float=None, color:str=None, chords:bool=False):
    """Only works for 2D"""
    if ax is None: fig, ax = plt.subplots()
    else: fig = None
    if scale is None: scale = 10 # factor to scale up displacements
    if color is None: color='red'
    line_objects = []
    plot_structure(state, ax)
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    ax.set_aspect('equal')

    if state.ndf > 2:
        n_curve=2   # number of plotting points along element

        for elem in state.elems:
            if len(elem.q) > 1:
                delta = np.array([[0.,0.],[0.,0.]])

                X = np.array([[elem.nodes[0].x, elem.nodes[0].y],
                            [elem.nodes[1].x, elem.nodes[1].y]])
                xl = np.linspace(0, elem.L, n_curve)
                print()
                xy = np.concatenate(([xl], [[-scale*elem.q["2"],scale*elem.q["3"]]]))
                xl, yl = elem.Rz_matrix()[:2,:2]@xy+[[elem.nodes[0].x]*n_curve,[elem.nodes[0].y]*n_curve]

                x0 = np.linspace(delta[0,0]*scale, delta[1,0]*scale, n_curve)
                y0 = np.linspace(delta[0,1]*scale, delta[1,1]*scale, n_curve)
                x = xl + x0
                y = yl + y0
                line_objects.extend(ax.plot(x, y, zorder = 1, color=color))
                # if abs(elem.cs) > 0.5:
                ax.fill_between(
                    [elem.nodes[0].x, elem.nodes[1].x],
                    [elem.nodes[0].y, elem.nodes[1].y],
                    y, color="r", alpha=0.2, interpolate=True)
                # else:
                #     plt.fill_between(
                #         [elem.nodes[0].x, elem.nodes[1].x],
                #         [elem.nodes[0].y, elem.nodes[1].y],
                #         y, color="r", alpha=0.2)

    # Plot undeformed chords
    n = 3 
    for elem in state.elems:
        x = np.linspace(elem.nodes[0].x, elem.nodes[1].x, n)
        y = np.linspace(elem.nodes[0].y, elem.nodes[1].y, n)
        ax.plot(x, y, **elem_style[0])

    f = 0.5
    for hinge in state.hinges:
        if hinge.node == hinge.elem.nodes[0]:
            x = hinge.node.x + f*hinge.elem.cs
            y = hinge.node.y + f*hinge.elem.sn
        else:
            x = hinge.node.x - f*hinge.elem.cs
            y = hinge.node.y - f*hinge.elem.sn
        ax.scatter(x, y, **hinge_style[0])
    return line_objects

def plot_U(model, U_vect, ax, scale=None, color=None, chords=False):
    """Only works for 2D"""
    
    if scale is None: scale = 10 # factor to scale up displacements
    if color is None: 
        color='red'
    A = mv.Kinematic_matrix(model)
    # print(A.f)
    U = U_vector(model, vector=U_vect)
    plot_structure(model, ax)
    for node in model.nodes:
        delta = [0.,0.]
        for i,dof in enumerate(node.dofs[0:2]):
            if not node.rxns[i]: 
                try: delta[i] = U[U.row_data.index(str(dof))]
                except: pass
        x = node.x
        y = node.y
        plt.plot( x, y, **node_style[0])
        plt.plot( x+delta[0]*scale, y+delta[1]*scale, color=color, **node_style[1])
        
    ###< Plot Chords
    # if chords:
    for elem in model.elems:
        X = np.array([[elem.nodes[0].x, elem.nodes[0].y], 
                    [elem.nodes[1].x, elem.nodes[1].y]])
        delta = np.array([[0.,0.],[0.,0.]])

        for j, node in enumerate(elem.nodes):
            for i,dof in enumerate(node.dofs[0:2]):
                if not node.rxns[i]: 
                    try:
                        delta[j,i] = U[U.row_data.index(str(dof))]
                    except:
                        pass
        x = np.linspace(X[0,0]+delta[0,0]*scale, X[1,0]+delta[1,0]*scale, 3)
        y = np.linspace(X[0,1]+delta[0,1]*scale, X[1,1]+delta[1,1]*scale, 3)
        plt.plot(x, y, ":", zorder = 1, color=color)
        
    # plot deformed curve
    if model.ndf > 2:
        V = A.c0@U.f # Element deformation vector
        n_curve=20   # number of plotting points along element

        for elem in model.elems:
            delta = np.array([[0.,0.],[0.,0.]])
            if hasattr(elem, "Elastic_curve"):
                X = np.array([[elem.nodes[0].x, elem.nodes[0].y], 
                              [elem.nodes[1].x, elem.nodes[1].y]])
                v_tags = [elem.tag+'_2', elem.tag+'_3']
                v = [V.get(v_tags[0]),V.get(v_tags[1])]
                xl = np.linspace(0, elem.L, n_curve)
                xl, yl = elem.Elastic_curve(xl, v, scale=scale, global_coord=True)
                

                for j, node in enumerate(elem.nodes): # Node displacements
                    for i,dof in enumerate(node.dofs[0:2]):
                        if not node.rxns[i]: # if there is no reaction at dof `i`...
                            try:
                                delta[j,i] = U[U.row_data.index(str(dof))]
                            except:
                                pass

                x0 = np.linspace(delta[0,0]*scale, delta[1,0]*scale, n_curve)
                y0 = np.linspace(delta[0,1]*scale, delta[1,1]*scale, n_curve)
                x = xl + x0
                y = yl + y0
                plt.plot(x, y, zorder = 1, color=color)

    # Plot undeformed chords
    n = 3 
    for elem in model.elems:
        x = np.linspace(elem.nodes[0].x, elem.nodes[1].x, n)
        y = np.linspace(elem.nodes[0].y, elem.nodes[1].y, n)
        plt.plot(x, y, **elem_style[0])

    f = 0.5
    for hinge in model.hinges:
        if hinge.node == hinge.elem.nodes[0]:
            x = hinge.node.x + f*hinge.elem.cs
            y = hinge.node.y + f*hinge.elem.sn
        else:
            x = hinge.node.x - f*hinge.elem.cs
            y = hinge.node.y - f*hinge.elem.sn
        plt.scatter(x, y, **hinge_style[0])

def plot_modes(model, shapes, ax, scale=None, color=None, label=None):
    """Only works for 2D"""
    if scale is None: scale = 5 # factor to scale up displacements
    if color is None: 
        color='red'
    A = mv.Kinematic_matrix(model)
    U = mv.Displacement_vector(A, shapes)
    X = []
    Y = []
    # plot_structure(model, ax)
    for node in model.nodes:
        delta = [0.,0.]
        for i,dof in enumerate(node.dofs[0:2]):
            if not node.rxns[i]: 
                try: delta[i] = U[U.row_data.index(str(dof))]
                except: pass
        x = node.x
        y = node.y
        X.append(x+delta[0]*scale)
        Y.append(y+delta[1]*scale)
        plt.plot( x, y, **node_style[0])
        # plt.plot( x+delta[0]*scale, y+delta[1]*scale, color=color, **node_style[1])

        
    ###< Plot Chords
    ne = len(model.elems)-1
    for i, elem in enumerate(model.elems):
        X = np.array([[elem.nodes[0].x, elem.nodes[0].y], 
                      [elem.nodes[1].x, elem.nodes[1].y]])
        delta = np.array([[0.,0.],[0.,0.]])

        for j, node in enumerate(elem.nodes):
            for i,dof in enumerate(node.dofs[0:2]):
                if not node.rxns[i]:
                    try:
                        delta[j,i] = U[U.row_data.index(str(dof))]
                    except:
                        pass

        x = np.linspace(X[0,0]+delta[0,0]*scale, X[1,0]+delta[1,0]*scale, 3)
        y = np.linspace(X[0,1]+delta[0,1]*scale, X[1,1]+delta[1,1]*scale, 3)

        plt.plot(X[:,0], X[:,1], **elem_style[0])
        if i==ne: 
            plt.plot(x, y, ':', zorder = 1, color=color, label=label)
        else:
            plt.plot(x, y, ':', zorder = 1, color=color)
    
    # plot deformed curve
    V = A.c0@U.f
    n_curve=20
    for elem in model.elems:
        delta = np.array([[0.,0.],[0.,0.]])
        if hasattr(elem, "Elastic_curve"):
            X = np.array([[elem.nodes[0].x, elem.nodes[0].y], 
                          [elem.nodes[1].x, elem.nodes[1].y]])
            v_tags = [elem.tag+'_2', elem.tag+'_3']
            v = [V.get(v_tags[0]),V.get(v_tags[1])]
            xl = np.linspace(0, elem.L, n_curve)
            xl, yl = elem.Elastic_curve(xl, v, scale=scale, global_coord=True)
            
            for j, node in enumerate(elem.nodes):
                for i,dof in enumerate(node.dofs[0:2]):
                    if not node.rxns[i]: 
                        try:
                            delta[j,i] = U[U.row_data.index(str(dof))]
                        except:
                            pass

            x0 = np.linspace(delta[0,0]*scale, delta[1,0]*scale, n_curve)
            y0 = np.linspace(delta[0,1]*scale, delta[1,1]*scale, n_curve)
            x = xl + x0
            y = yl + y0
            plt.plot(x, y, zorder = 1, color=color)
    
def plot_structure(Model, ax=None, label=False):
    if ax is None: _,ax = plt.subplots()
    n =3
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    ax.set_aspect('equal')

    ###< Plot Elements
    f = 0.7 # parameter controling element label distance from element
    for elem in Model.elems:
        x = np.linspace(elem.nodes[0].x0, elem.nodes[1].x0, n)
        y = np.linspace(elem.nodes[0].y0, elem.nodes[1].y0, n)
        # plt.plot(x, y, zorder = 1, color ='grey')
        ax.plot(x, y, **elem_style[0])
        if label:
            # label element tag
            xl = x[1] - elem.sn*f
            yl = y[1] - elem.cs*f
            ax.annotate("${}$".format(elem.tag), xy=(xl, yl))

    f = 0.4
    if Model.ndf == 3:
        for elem in Model.elems:
            if type(elem) is Truss:
                x = elem.nodes[0].x0 + f*elem.cs
                y = elem.nodes[0].y0 + f*elem.sn
                ax.scatter(x, y, s=20, zorder=2, facecolors='white', edgecolors='black')
                x = elem.nodes[1].x0 - f*elem.cs
                y = elem.nodes[1].y0 - f*elem.sn
                ax.scatter(x, y, s=20, zorder=2, facecolors='white', edgecolors='black')

    ###>
    # show hinges
    f = 0.5
    for hinge in Model.hinges:
        if hinge.node == hinge.elem.nodes[0]:
            x = hinge.node.x + f*hinge.elem.cs
            y = hinge.node.y + f*hinge.elem.sn
        else:
            x = hinge.node.x - f*hinge.elem.cs
            y = hinge.node.y - f*hinge.elem.sn
        ax.scatter(x, y, **hinge_style[0])

    # show reactions
    offset = 0.5
    rx_offset = [offset, 0., 0.]
    ry_offset = [0., offset, 0.]
    for rxn in Model.rxns:
        if rxn.node.rxns == [1, 1, 0]: # pinned reaction
            pass
            #plt.scatter(x, y, s=25, zorder=2, color='black', marker = '^')
        elif rxn.node.rxns == [1, 0, 0]: # x - roller
            pass
            #plt.scatter(x, y, s=25, zorder=2, color='black', marker = 'o')
        elif rxn.node.rxns == [0, 1, 0]: # y - roller
            pass
        elif rxn.node.rxns == [1, 1, 1]: # fixed reaction
            pass
        
    for node in Model.nodes:
        for i, rxn in enumerate(node.rxns):
            if rxn:
                x = node.x0-rx_offset[i]#*np.sign(node.elems[0].cs)
                y = node.y0-ry_offset[i]#*np.sign(node.elems[0].sn)
                ax.plot(x, y, **rxn_style[i])

    # plot nodes
    f = 0.4 # factor to tweak annotation distance
    for node in Model.nodes:
        ax.plot(node.x,node.y, color = 'black', marker = 's')
        if label:
            if sum(node.rxns) == 0:
                ax.annotate(node.tag, xy=(f+node.x, 0.5*f+node.y), zorder = 3, color = 'blue')
            else:
                tag = node.tag + " "+ str(node.rxns)
                ax.annotate(tag, xy=(f+node.x, 0.5*f+node.y), zorder = 3, color = 'blue')
    return ax

def plot_beam(Model, ax,label=False):
    n =3
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    # ax.set_aspect('equal')

    ###< Plot Elements
    f = 0.5 # parameter controling element label distance from element
    for elem in Model.elems:
        x = np.linspace(elem.nodes[0].x, elem.nodes[1].x, n)
        y = np.linspace(0.0, 0.0, n)
        plt.plot(x, y, zorder = 1, color ='grey')
        # label element tag
        xl = (x[0] + x[-1])/2
        yl = 0.0002
        if label:
            ax.annotate(elem.tag, xy=(xl, yl))

    # show hinges
    f = 0.1
    for hinge in Model.hinges:
        if hinge.node == hinge.elem.nodes[0]:
            x = hinge.node.x + f*hinge.elem.L
            y = 0.0 
        else:
            x = hinge.node.x - f*hinge.elem.L
            y = 0.0 

        plt.scatter(x, y, s=20, zorder=2, facecolors = 'white', edgecolors='black')

    # plot nodes
    y_off = 0.0004 # factor to tweak annotation distance
    for node in Model.nodes:
        plt.plot(node.x,0.0, color = 'black', marker = 's')
        if sum(node.rxns) == 0:
            ax.annotate(node.tag, xy=(node.x, y_off), zorder = 3, color = 'blue')
        else:
            tag = node.tag + " "+ str(node.rxns)
            if label:
                ax.annotate(tag, xy=(node.x, y_off), zorder = 3, color = 'blue')

def plot_structure3d(Model, ax):
    n =3

    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')

        ###< Plot Elements
    f = 0.9 # parameter controling element label distance from element
    for elem in Model.elems:
        x = np.linspace(elem.nodes[0].x, elem.nodes[1].x, n)
        y = np.linspace(elem.nodes[0].y, elem.nodes[1].y, n)
        z = np.linspace(elem.nodes[0].z, elem.nodes[1].z, n)
        xl = x[1] - elem.sn*f
        yl = y[1]- elem.cs*f
        zl = z[1] - elem.cz*f
        ax.plot(x, y, z, color ='black')
 #         ax.annotate(elem.tag, xyz = (xl, yl, zl))
        ax.text(xl, yl, zl, elem.tag, color='red')
        # plot nodes
    f = 0.4 # factor to tweak annotation distance
    for node in Model.nodes:
        ax.scatter(node.x,node.y, node.z, color = 'black', marker = 's')
 #             if sum(node.rxns) == 0:

 #                 ax.annotate(node.tag, xy=(f+node.x, 0.5*f+node.y), zorder = 3, color = 'blue')
 #             else:
 #                 tag = node.tag + " "+ str(node.rxns)
 #                 ax.annotate(tag, xy=(f+node.x, 0.5*f+node.y), zorder = 3, color = 'blue')
    return ax

def plot_2dshape(xyz,N,node=1,ax=None):
    if ax is None: _,ax = plt.subplots(1,1,subplot_kw={'projection':'3d'})

    xx, yy = np.meshgrid(xyz[:,0],xyz[:,1])
    z = N[(i-1)*2,0](xx,yy)










from IPython.display import display_html

def disp_sbs(*args):
    html_str = ''
    for df in args:
        html_str += df.to_html()
    display_html(html_str.replace('table', 'table style = "display:inline"'), raw=True)